# statement.py
# Copyright 2017 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Chess Query Language (CQL) statement parser.

The basic structure of a CQL statement at version 5.1 is:

'cql ( parameters ) list_of_filters'.

This module allows both parameters and list_of_filters to be empty.

The keywords 'or', 'not', and '{ ... }', alter the default implied 'and' between
adjacent filters in the list.  All three keywords generate a filter which takes
it's place in a list_of_filters.

A number of named operations may be applied to a filter in the form:

'operation optional_range filter' where filter is something that can appear as
a single element in a list_of_filters.  Operations can be chained:

'operation1 optional_range1 operation2 opetional_range2 ... filter', and the
generated filter appears as a single element in a list of filters.

A number of named operations take arguments in the form '( arguments )'. Some
arguments are list_of_filters.

"""

import re
import collections

import pgn_read.core.constants

from . import constants, node

CQL_PARAMETERS = (constants.OUTPUT,
                  constants.INPUT,
                  constants.GAMENUMBER,
                  constants.YEAR,
                  constants.SILENT,
                  constants.PLAYER,
                  constants.ELO,
                  constants.SITE,
                  constants.EVENT,
                  constants.MATCHCOUNT,
                  constants.SORT,
                  )

# ATTACK is a shorthand for:
# 'ray orthogonal ( [QqRr] on s, ... ) or ray diagonal ( [QqBb] on s, ... )
# and is included as a RAY_DIRECTION.
# ATTACK is not the 'attack' filter in this context.
RAY_DIRECTIONS = frozenset(
    constants.DIRECTION_FILTER.union((constants.ATTACK,)))

TRANSFORM_FILTERS = frozenset((constants.FLIP,
                               constants.FLIPDIHEDRAL,
                               constants.FLIPHORIZONTAL,
                               constants.FLIPVERTICAL,
                               constants.FLIPCOLOR,
                               constants.ROTATE45,
                               constants.ROTATE90,
                               constants.SHIFT,
                               constants.SHIFTHORIZONTAL,
                               constants.SHIFTVERTICAL,
                               ))
STAR_FILTERS = frozenset((constants.NEXT_STAR,
                          constants.PREVIOUS_STAR,
                          ))

# MAINLINE, PREVIOUS, and NEXT, are move modifiers not filters in this context.
MOVE_MODIFIERS = frozenset((constants.MAINLINE,
                            constants.PREVIOUS,
                            constants.NEXT,
                            constants.EMPTY,
                            ))

RELATION_ECHO_PARAMETERS = frozenset((constants.ECHOFLIP,
                                      constants.ECHOFLIPVERTICAL,
                                      constants.ECHOFLIPHORIZONTAL,
                                      constants.ECHOSHIFT,
                                      constants.ECHOSHIFTHORIZONTAL,
                                      constants.ECHOSHIFTVERTICAL,
                                      constants.ECHOROTATE90,
                                      ))
RELATION_TOMOVE_ARGUMENTS = frozenset((constants.MATCH,
                                       constants.MISMATCH,
                                       ))
RELATION_LCA_PLAIN_PARAMETERS = frozenset((constants.ANCESTOR,
                                           constants.DESCENDANT,
                                           ))
RELATION_LCA_RANGE_PARAMETERS = frozenset((constants.LCAMAX,
                                           constants.LCASOURCE,
                                           constants.LCASUBSTRING,
                                           constants.LCASUM,
                                           constants.LCATARGET,
                                           ))
RELATION_SQUARE_PARAMETERS = frozenset((constants.MATCH,
                                        constants.MISMATCH,
                                        constants.SOURCESQUARES,
                                        constants.TARGETSQUARES,
                                        ))
GAME_RESULTS = (
    pgn_read.core.constants.WHITE_WIN,
    pgn_read.core.constants.BLACK_WIN,
    pgn_read.core.constants.DRAW)
REAL_BRACES = frozenset((constants.CQL,
                         constants.LEFT_BRACE_FILTER,
                         constants.LEFT_PARENTHESIS_FILTER,
                         constants.RELATION,
                         ))
IMPLIED_BRACES = frozenset((constants.CQL,
                            constants.LEFT_BRACE_FILTER,
                            constants.LEFT_PARENTHESIS_FILTER,
                            constants.RELATION,
                            constants.NOT,
                            constants.OR,
                            constants.ON,
                            constants.TO,
                            constants.FROM,
                            constants.IN,
                            ))
PARAMETER_BRACES = frozenset((constants.OR,
                              constants.ON,
                              constants.TO,
                              constants.FROM,
                              constants.IN,
                              ))
VARIABLE_TYPES = frozenset((constants.PIECE, constants.SQUARE))

# Better name needed to indicate shared syntax.
RAY_FILTERS = frozenset(
    RAY_DIRECTIONS.union(TRANSFORM_FILTERS.union(STAR_FILTERS)))

COUNTABLE_FILTERS = frozenset()
COUNTABLE_RANGE_FILTERS = frozenset((constants.ATTACK,
                                     constants.COUNTSQUARES,
                                     constants.SQUARE,
                                     constants.NEXT,
                                     constants.PREVIOUS,
                                     constants.NEXT2,
                                     constants.PREVIOUS2,
                                     constants.NEXT_STAR,
                                     constants.PREVIOUS_STAR,
                                     constants.POWER,
                                     constants.POWERDIFFERENCE,
                                     constants.MOVENUMBER,
                                     constants.YEAR,
                                     constants.ELO,
                                     constants.RAY,
                                     constants.MATCHCOUNT,
                                     ))

Token = collections.namedtuple('Token', constants.TOKEN_NAMES)

_ROTATE45_VALIDATION_TABLE = str.maketrans('12345678', 'xxxxxxxx')


class StatementError(Exception):
    pass


class Statement:
    """CQL statement parser for version 5.1.

    Parse text for a CQL statement.
    
    """
    create_node = node.Node

    def __init__(self):
        """"""
        super().__init__()
        self._description_string = ''
        self._statement_string = ''

        # Error handling is slightly different to .querystatement module.
        # The error information object is held directly here while it is held
        # in the Where instance used to calculate the query there.
        # This module processes instructions used to generate Where instances.
        self._reset_state()
        self._error_information = None

    def _reset_state(self):
        """Initialiase the CQL parsing state.

        The analyser is able to have at least two goes at a statement.

        """
        self._error_information = False
        self.tokens = None
        self.cql_tokens = []
        self.stack = []
        self.node_stack = []
        self.brace_parenthesis_stack = []
        self.cql_parameters = None
        self.cql_filters = None

        # The scoping rules for piece and square variables at cql5.1 seem to be:
        # $name is unique across piece and square use and cannot be redefined.
        # $name is visible in all enclosed scopes.
        # $name scope is bounded by '( ... )' and '{ ... }'.
        # This is not certain because:
        # '{square $x in h1-8 attack (R $x)} attack (Q $x)' gives error unbound,
        # 'square $x in h1-8 attack (R $x) attack (Q $x)' is ok,
        # '{shift 1 square $x in h1-8 attack (R $x)} attack (Q $x)' is ok.
        # However:
        # '{shift 0 square $x in h1-8 attack (R $x)} attack (Q $x)' gives
        # error undeclared so 'attack (Q $x)' is probably ignored in the
        # 'shift 1' version because the shift filter evaluates False or empty,
        # so the scoping assumption is probably sound.

        # All declared variable names.
        self.variables = {}

        # Variable names declared in scope.
        self.variables_stack = []

    @property
    def cql_error(self):
        return self._error_information

    @cql_error.setter
    def cql_error(self, value):
        if self._error_information:
            return
        self._error_information = ErrorInformation(
            self._statement_string.strip())
        self._error_information.description = value

    def get_name_text(self):
        """Return name text."""
        return self._description_string

    def get_name_statement_text(self):
        """Return name and statement text."""
        return constants.NAME_DELIMITER.join(
            (self._description_string,
             self._statement_string,
             ))

    def get_statement_text(self):
        """Return statement text."""
        return self._statement_string
        
    def lex(self):
        """Split the ChessQL statement into tokens."""
        self.tokens = re.findall(constants.CQL_PATTERN,
                                 self._statement_string)
        if self._statement_string.strip() == '':
            self._error_information = False
            return
        
    def parse(self):
        """Generate a tree from tokens describing the query."""

        # lex() sets self._error_information False if statement is worth
        # parsing
        if self._error_information:
            return
        
        self.stack.append(self.look_for_cql_parameters_or_body)
        while len(self.tokens):

            # 'bool(self.stack[-1]()) == True' means an error is detected and
            # an error reporter has been appended to self.stack.
            if self.stack[-1]():
                self.stack[-1]()
                break

        else:
            self.stack[-1]()

    def validate(self):
        """Return None if query is valid, or a WhereStatementError instance."""

        if self._error_information:
            return self._error_information
        if not self.cql_filters:
            return None
        return self._rotate45_specific_squares(self.cql_filters, [])
        
    def _rotate45_specific_squares(self, filter_, rotate45stack):
        if filter_.type == constants.ROTATE45:
            rotate45stack.append(None)
        for n in filter_.children:
            if rotate45stack and n.type == constants.PIECE_DESIGNATOR_FILTER:
                #if {c for c in n.leaf}.intersection('12345678'))
                if n.leaf != n.leaf.translate(_ROTATE45_VALIDATION_TABLE):
                    self._error_information = ErrorInformation(
                        self._statement_string.strip())
                    self._error_information.description = (
                        'rotate45 on specific squares')
                    return self._error_information
            r45ss = self._rotate45_specific_squares(n, rotate45stack)
            if r45ss:
                return r45ss
        if filter_.type == constants.ROTATE45:
            rotate45stack.pop()
        return None
        
    def process_statement(self, text):
        """Lex and parse the ChessQL statement."""

        # Assume no error, but set False indicating process_query_statement
        # has been called.
        self._error_information = False

        # First attempt treats whole text as a ChessQL statement.
        # Second attempt treats first line, of at least two, as the query's
        # name and the rest as a ChessQL statement.
        for rule in (('', text.strip()),
                     [t.strip()
                      for t in text.split(constants.NAME_DELIMITER, 1)]):
            
            if len(rule) == 1:

                # The second element of rule is being processed and the text in
                # rule[1] cannot be a valid ChessQL statement because it is the
                # text which was processed from the first element of rule.
                return False

            self._reset_state()
            self._description_string, self._statement_string = rule
            self.lex()
            self.parse()
            if self.validate():
                continue
            return True

        return None

    def is_statement(self):
        return not self._error_information

    def p_error(self):
        self._error_information = ErrorInformation(
            self._statement_string.strip())
        self._error_information.tokens = self.cql_tokens.copy()
        if self.tokens:
            self._error_information.next_token = self.tokens[0]
        return True

    def look_for_cql_parameters_or_body(self):
        if self.tokens:
            if self.cql_parameters is None:
                self.stack.append(self.p_cql)
                return
            if self.cql_filters is None:
                self.stack.append(self.b_cql)
                return
            self.p_error()
            return True

    def pop_top_of_stack(self):
        self.node_stack.pop()
        self.stack.pop()

    def p_numbers(self, stackexit=None):
        # Minimum and maximum are held in the range attribute.  In most cases
        # numbers are applied to the following filter, but year and elo are
        # notable exceptions which need not be processed by p_range().
        ns = self.node_stack
        token = Token(*self.tokens[0])
        if not token.number:
            if stackexit is None:
                self.pop_top_of_stack()
            else:
                self.stack[-1] = stackexit
            return False
        if len(ns[-1].range) > 1:
            self.stack.append(self.p_error)
            return True
        ns[-1].range.append(token.number)
        self.cql_tokens.append(self.tokens.pop(0))

    def b_numbers(self):
        if not self.tokens:
            self.b_end()
            return
        return self.p_numbers()

    def p_range(self, stackexit=None):
        # When processing a number as part of a range there must be at least
        # one more token: the object of the range.
        if self.p_numbers(stackexit=stackexit):
            return True
        if not len(self.tokens):
            self.stack.append(self.p_error)
            return True

    def b_range(self):
        # Hack so '... shift 8 left b ...' and similar come out correct.
        return self.p_range(stackexit=self.b_body)

    # p_* methods deal with 'cql(...)' part of 'cql(...)...' statement.
    # b_* methods deal with final '...' part of 'cql(...)...' statement .

    def p_cql(self):
        token = Token(*self.tokens[0])
        if token.cql != constants.CQL:
            self.stack.append(self.p_error)
            return True
        self.node_stack.append(self.create_node(token.cql))
        self.stack.append(self.p_left_parenthesis)
        self.cql_tokens.append(self.tokens.pop(0))

    def p_left_parenthesis(self):
        token = Token(*self.tokens[0])
        if token.left_parenthesis == constants.LEFT_PARENTHESIS_FILTER:
            self.add_argument(token.left_parenthesis, self.p_cql_parameter)
        else:
            self.stack.append(self.p_error)
            return True

    def p_cql_parameter(self):
        ns = self.node_stack
        token = Token(*self.tokens[0])
        if token.cql_parameter:
            if ns[-1].repeated_parameter(token.cql_parameter):
                self.stack.append(self.p_error)
                return True
            ns[-1].children.append(self.create_node(token.cql_parameter))
            ns.append(ns[-1].children[-1])
            if token.cql_parameter in (constants.INPUT, constants.OUTPUT):
                self.stack.append(self.p_allowed_strings_filename)
            elif token.cql_parameter == constants.VARIATIONS:
                ns[-1].leaf = True
                ns.pop() # Cancel the ns.append(...) before 'if' clause.
            elif token.cql_parameter == constants.GAMENUMBER:
                self.stack.append(self.p_gamenumber)
                ns[-1].leaf = ns[-1].range
            elif token.cql_parameter == constants.MATCHCOUNT:
                self.stack.append(self.p_matchcount)
                ns[-1].leaf = ns[-1].range
            else:
                self.stack.append(self.p_error)
                return True
            self.cql_tokens.append(self.tokens.pop(0))
        elif token.site_event:

            # Multiple event or site parameters are accepted by cql5.1, but
            # all of each must match.
            #if ns[-1].repeated_parameter(token.site_event):
            #    self.stack.append(self.p_error)
            #    return True
            
            self.add_argument(token.site_event, self.p_double_quoted_string)
        elif token.player:
            ns.append(ns[-1].get_repeatable_parameter(token.player,
                                                      self.create_node))
            self.stack.append(self.p_player_white_black)
            self.cql_tokens.append(self.tokens.pop(0))
        elif token.elo:
            ns.append(ns[-1].get_repeatable_parameter(token.elo,
                                                      self.create_node))
            self.stack.append(self.p_elo_white_black)
            self.cql_tokens.append(self.tokens.pop(0))
        elif token.year:

            # Multiple year parameters are accepted by cql5.1, but all years
            # must match.
            #if ns[-1].repeated_parameter(token.year):
            #    self.stack.append(self.p_error)
            #    return True
            
            self.add_argument(token.year, self.p_range)
            ns[-1].leaf = ns[-1].range
        elif token.silent:

            # Multiple silent parameters are accepted but it is not clear if
            # instruction is tied to all filters or just to next parameter.
            # Apply to all seems most likely but why allow multiple silent
            # parameters when all other parameters are restricted to a single
            # occurrence.
            #if ns[-1].repeated_parameter(token.silent):
            #    self.stack.append(self.p_error)
            #    return True
            
            self.add_argument_leaf(token.silent, True)
        elif token.sort:
            if ns[-1].repeated_parameter(token.sort):
                self.stack.append(self.p_error)
                return True
            self.add_argument(token.sort, self.p_sort_matchcount)
        elif token.result:

            # Multiple result parameters are accepted by cql5.1 but all results
            # must match.
            #if ns[-1].repeated_parameter(token.result):
            #    self.stack.append(self.p_error)
            #    return True

            self.add_argument(token.result, self.p_allowed_strings_result)
        elif token.right_parenthesis == constants.RIGHT_PARENTHESIS_TYPE:
            self.pop_top_of_stack()
            if len(ns) != 1:
                self.stack.append(self.p_error)
                return True
            if ns[-1].type != constants.CQL:
                self.stack.append(self.p_error)
                return True
            self.cql_parameters = ns[-1]
            self.pop_top_of_stack()
            self.stack.pop()
            self.cql_tokens.append(self.tokens.pop(0))
        else:
            self.stack.append(self.p_error)
            return True

    def p_allowed_strings_filename(self):
        token = Token(*self.tokens[0])
        if not token.allowed_strings.endswith(constants.PGN_FILE_VALUE):
            self.stack.append(self.p_error)
            return True
        self.node_stack[-1].leaf = token.allowed_strings
        self.pop_top_of_stack()
        self.cql_tokens.append(self.tokens.pop(0))

    def p_allowed_strings_result(self):
        token = Token(*self.tokens[0])
        if token.allowed_strings.endswith(constants.PGN_FILE_VALUE):
            self.stack.append(self.p_error)
            return True
        self.node_stack[-1].leaf = token.allowed_strings
        self.pop_top_of_stack()
        self.cql_tokens.append(self.tokens.pop(0))

    def b_allowed_strings_result(self):
        if not self.tokens:
            self.b_end()
            return
        return self.p_allowed_strings_result()

    def b_allowed_strings_flipcolor_result(self):
        if not self.tokens:
            self.b_end()
            return
        token = Token(*self.tokens[0])
        if token.allowed_strings.endswith(constants.PGN_FILE_VALUE):
            self.stack.append(self.p_error)
            return True
        self.node_stack[-1].leaf = token.allowed_strings
        self.pop_top_of_stack()

        # Extra pop-op, compared with p_allowed_strings_result to reverse the
        # push-op done for flipcolor.  Now looks like a mistake to press
        # p_allowed_strings_result into all three cases.
        self.pop_top_of_stack()

        self.cql_tokens.append(self.tokens.pop(0))

    def p_gamenumber(self):
        token = Token(*self.tokens[0])
        if not token.number:
            self.stack.append(self.p_error)
            return True
        self.stack[-1] = self.p_range

    def p_matchcount(self):
        token = Token(*self.tokens[0])
        if not token.number:
            self.stack.append(self.p_error)
            return True
        self.stack[-1] = self.p_range

    def p_double_quoted_string(self):
        token = Token(*self.tokens[0])
        if not token.double_quoted_string:
            self.stack.append(self.p_error)
            return True
        self.node_stack[-1].leaf = token.double_quoted_string
        self.pop_top_of_stack()
        self.cql_tokens.append(self.tokens.pop(0))

    def b_double_quoted_string(self):
        if not self.tokens:
            self.b_end()
            return
        return self.p_double_quoted_string()

    def p_sort_matchcount(self):
        token = Token(*self.tokens[0])
        if not token.cql_parameter == constants.MATCHCOUNT:
            self.stack.append(self.p_error)
            return True
        self.node_stack[-1].leaf = True
        self.pop_top_of_stack()

    # Differs from b_player_white_black due to rules for mixing parameters.
    def p_player_white_black(self):
        ns = self.node_stack
        token = Token(*self.tokens[0])
        if token.white_black_keywords:
            if (ns[-1].repeated_parameter(token.white_black_keywords) or
                ns[-1].leaf
                ):
                self.stack.append(self.p_error)
                return True
            self.add_argument(token.white_black_keywords,
                              self.p_double_quoted_string)
        elif token.double_quoted_string:
            if len(ns[-1].children):
                self.stack.append(self.p_error)
                return True
            if ns[-1].leaf:
                self.stack.append(self.p_error)
                return True
            ns.append(ns[-1])
            self.stack.append(self.p_double_quoted_string)
        elif ns[-1].type == constants.PLAYER:
            self.pop_top_of_stack()
        else:
            self.stack.append(self.p_error)
            return True

    # Differs from p_elo_white_black_body following rules for mixing parameters.
    # Uses p_range rather than p_numbers.
    def p_elo_white_black(self):
        ns = self.node_stack
        token = Token(*self.tokens[0])
        if token.white_black_keywords:
            if (ns[-1].repeated_parameter(token.white_black_keywords) or
                ns[-1].range
                ):
                self.stack.append(self.p_error)
                return True
            self.add_argument(token.white_black_keywords, self.p_numbers)
            ns[-1].leaf = ns[-1].range
        elif token.number:
            if len(ns[-1].children):
                self.stack.append(self.p_error)
                return True
            if ns[-1].range:
                self.stack.append(self.p_error)
                return True
            ns.append(ns[-1])
            self.stack.append(self.p_range)
            ns[-1].leaf = ns[-1].range
        elif ns[-1].type == constants.ELO:
            self.pop_top_of_stack()
        else:
            self.stack.append(self.p_error)
            return True

    # p_* methods deal with 'cql(...)' part of 'cql(...)...' statement.
    # b_* methods deal with final '...' part of 'cql(...)...' statement .

    def b_pop_top_of_stack(self):
        self.pop_top_of_stack()

    def b_pop_nested_stack(self):
        while True:
            if self.node_stack[-1].type in REAL_BRACES:
                break
            self.node_stack[-1].derive_filter_type_from_children()
            self.b_pop_top_of_stack()

    def b_pop_nested_stack_to_implied_brace(self):
        while True:
            if self.node_stack[-1].type in IMPLIED_BRACES:
                break
            self.node_stack[-1].derive_filter_type_from_children()
            self.b_pop_top_of_stack()

    def b_pop_top_of_stack_and_nested(self):

        # Caller is responsible for setting self.node_stack[-1].setfilter to
        # fit b_pop_top_of_stack. 
        self.b_pop_top_of_stack()
        self.b_pop_nested_stack()

    def b_pop_variables_stack(self):
        for v in self.variables_stack[-1]:
            del self.variables[v]
        self.variables_stack.pop()

    # Intend to use this to frame square and piece filters
    # square variable_name in in_set filter1 filter2 ...
    # For example
    # shift square $x in h1-8 attack ( R $x ) attack ( $x K )
    # is different to
    # shift { square $x in h1-8 attack ( R $x ) } attack ( $x K )
    # because the scope of square is limited by {} in the second version.
    def b_cql(self):
        self.node_stack.append(self.create_node(constants.CQL))
        self.stack.append(self.b_body)
        self.variables_stack.append(set())

    def b_body(self):
        if not self.tokens:
            self.b_end()
            return
        token = Token(*self.tokens[0])
        self.b_collapse_filters_stack()
        return self.add_cql_body(token, self.b_body)

    def b_piece_square_in_set_body(self):
        if not self.tokens:
            self.b_end()
            return
        token = Token(*self.tokens[0])
        self.b_collapse_filters_stack()
        return self.add_piece_square_in_set_body(
            token, self.b_piece_square_in_set_body)

    def b_collapse_filters_stack(self):
        # Collapse stack at end of filters chain before processing token.
        # Originally first thing in add_cql_body() but needs to be done before
        # this and add_set_body() and add_piece_square_in_set_body() calls.
        if self.node_stack[-1].type in RAY_FILTERS:
            self.stack[-1] = self.b_pop_top_of_stack

    def b_collapse_parameters_stack(self):
        # Collapse stack at end of 'or' and 'on' sequences after parameters
        # such as 'from' and 'to'
        if not self.tokens:
            return
        token = Token(*self.tokens[0])
        while True:
            if self.node_stack[-1].type not in PARAMETER_BRACES:
                break
            if token.or_ or token.on:
                break
            self.b_pop_top_of_stack()

    def b_end(self):
        # Collapse stack for 'cql() not { ... }' and similar: 'not' is left on
        # stack at end of tokens in this example.
        self.b_pop_nested_stack()
        if len(self.node_stack) == 1:
            self.cql_filters = self.node_stack[-1]
            self.node_stack.pop()
            self.variables_stack.pop()

    def add_piece_square_in_set_body(self, token, b_method):
        # square and piece excluded because statements like:
        # 'square $x in square y$ in ...' are not allowed.
        if token.piece_designator:
            self.add_piece_designator(token)
        elif token.plain:
            if token.plain == constants.DARKSQUARES:
                self.add_plain_filter(token)
            elif token.plain == constants.LIGHTSQUARES:
                self.add_plain_filter(token)
            else:
                self.stack.append(self.p_error)
                return True
        elif token.not_:
            self.add_argument(token.not_, self.b_body)
        elif token.or_:
            if self.add_binary_argument(token.or_, b_method):
                return True
        elif token.on:
            if self.add_binary_argument(token.on, b_method):
                return True
        elif token.transform: # includes direction filters
            self.add_argument(token.transform, self.b_body)
        elif token.attack:
            self.add_argument(token.attack, self.b_attack)
        elif token.move:
            self.add_argument(token.move, self.b_move)
        elif token.between: # not listed as a set filter, but described so.
            self.add_argument(token.between, self.b_between)
        elif token.left_brace:
            self.variables_stack.append(set())
            self.add_argument(token.left_brace, self.b_body)
        elif token.right_brace:
            node_stack = self.node_stack
            stack = self.stack
            variables = self.variables
            variables_stack = self.variables_stack
            while True:
                if not node_stack:
                    stack.append(self.p_error)
                    return True
                elif node_stack[-1].type == constants.LEFT_BRACE_FILTER:
                    node_stack[-1].derive_brace_filter_type_from_children()
                    for v in variables_stack[-1]:
                        del variables[v]
                    variables_stack.pop()
                    node_stack.pop()
                    stack.pop()
                    break
                elif node_stack[-1].type == constants.CQL:
                    node_stack[-1].derive_brace_filter_type_from_children()
                    for v in variables_stack[-1]:
                        del variables[v]
                    variables_stack.pop()
                    node_stack.pop()
                    stack.pop()
                elif node_stack[-1].type == constants.NOT:
                    node_stack[-1].derive_brace_filter_type_from_children()
                    node_stack.pop()
                    stack.pop()
                else:
                    stack.append(self.p_error)
                    return True
            self.cql_tokens.append(self.tokens.pop(0))
            if len(self.tokens) == 0:
                self.b_pop_nested_stack()
        else:
            self.stack.append(self.p_error)
            return True

    def add_set_body(self, token, b_method):

        # Do not delegate plain filter to add_piece_square_in_set_body
        if token.plain:
            self.add_plain_filter(token)

        elif token.piece_square_variable:
            return self.add_piece_square_variable(token)
        else:
            return self.add_piece_square_in_set_body(token, b_method)

    def add_cql_body(self, token, b_method):
        if token.piece_square:
            self.add_argument(token.piece_square, self.b_piece_square_filter)
        elif token.number:
            self.node_stack.append(self.node_stack[-1])
            self.stack.append(self.b_range) # Hides a hack which needs fixing.
        elif token.countsquares_power:
            self.add_argument(token.countsquares_power,
                              self.b_countsquares_power)
        elif token.powerdifference:
            self.add_argument(token.powerdifference, self.b_powerdifference)
        elif token.next_previous:
            self.add_argument(token.next_previous, self.b_next_previous)
        elif token.ray:
            self.add_argument(token.ray, self.b_ray)
        elif token.relation:
            self.add_argument(token.relation, self.b_relation)
        elif token.elo:
            self.add_argument(token.elo, self.b_elo_white_black)
        elif token.player:
            self.add_argument(token.player, self.b_player_white_black)
        elif token.site_event:
            self.add_argument(token.site_event, self.b_double_quoted_string)
        elif token.result:
            if self.node_stack[-1].type == constants.FLIPCOLOR:
                self.add_argument(token.result,
                                  self.b_allowed_strings_flipcolor_result)
            else:
                self.add_argument(token.result, self.b_allowed_strings_result)
        elif token.year:
            self.add_argument(token.year, self.b_numbers)
            self.node_stack[-1].leaf = self.node_stack[-1].range
        elif token.cql_parameter == constants.GAMENUMBER:
            self.add_argument(token.cql_parameter, self.b_numbers)
            self.node_stack[-1].leaf = self.node_stack[-1].range
        elif token.mainline:
            self.add_argument_leaf(constants.PLAIN_FILTER, token.mainline)
            self.b_pop_nested_stack_to_implied_brace()
            self.b_collapse_parameters_stack()
        elif token.hascomment:
            self.add_argument(token.hascomment, self.b_double_quoted_string)
        elif token.silent:
            if len(self.tokens) < 2:
                self.stack.append(self.p_error)
                return True
            t = Token(*self.tokens[1])
            if t.next_previous or t.relation:
                self.add_argument_leaf(token.silent, True)
            else:
                self.stack.append(self.p_error)
                return True

        # Incomplete because there is no list of filters which are countable
        # without a range: there are some because the possibility is mentioned,
        # and the filter's documentation will say.
        #
        # Syntax of sort appears to be, with 'max' as default:
        # sort [max|min] ["Documentation string"] <countable filter>.
        #
        elif token.sort:
            if len(self.tokens) < 2:
                self.stack.append(self.p_error)
                return True
            t = ''.join(Token(*self.tokens[1]))
            if t in COUNTABLE_FILTERS:
                self.add_argument_leaf(token.sort, True)
            elif t in COUNTABLE_RANGE_FILTERS:
                self.add_argument_leaf(token.sort, True)
            else:
                self.stack.append(self.p_error)
                return True

        # Some set filters may not be such if qualified by a range.  This is
        # acceptable because the point is reliably identifying tokens which
        # can be ignored: if it turns out that 'comment' should not have been
        # used it does not matter because comment filters are ignored.
        elif token.comment:
            if len(self.tokens) < 2:
                self.stack.append(self.p_error)
                return True
            t = Token(*self.tokens[1])
            if t.double_quoted_string:
                self.add_argument_leaf(token.comment, t.double_quoted_string)
                self.cql_tokens.append(self.tokens.pop(0))
            elif t.piece_designator:
                self.add_argument_leaf(token.comment, True)

            # Default action for any other set filter.
            # Dihedral and shift transforms have been added to SET_FILTERS
            # since this was written: it may be incorrect now.
            elif ''.join(t) in constants.SET_FILTER:
                self.add_argument_leaf(token.comment, True)

            else:
                self.stack.append(self.p_error)
                return True

        else:
            return self.add_set_body(token, b_method)

    def add_re_cql_body(self, token, b_method):
        #
        # Copied from http://www.gadycosteff.com/cql/doc/next.html (cql5.1)
        #
        # next-filter := 'next' opt-range '(' constituent+ ')'
        # previous-filter := 'previous' opt-range '(' constituent+ ')'
        # constituent := filter
        #              | filter '*'
        #              | filter '+'
        #              | filter '?'
        #              | '(' filter-list ')'
        #
        # next2 and previous2 have the same syntax.
        #
        if token.left_parenthesis:
            self.add_left_parenthesis(token)
            self.stack.append(self.b_one_or_more_filters)
        elif token.repeat_operators:
            children = self.node_stack[-1].children
            if not children:
                self.stack.append(self.p_error)
                return True
            if children[-1].repeat:
                self.stack.append(self.p_error)
                return True
            if children[-1].type == constants.LEFT_PARENTHESIS_FILTER:
                self.stack.append(self.p_error)
                return True
            children[-1].repeat = token.repeat_operators
            self.cql_tokens.append(self.tokens.pop(0))
        else:
            return self.add_cql_body(token, b_method)

    def add_piece_designator(self, token):
        c = self.node_stack[-1].children
        c.append(self.create_node(constants.PIECE_DESIGNATOR_FILTER))
        c[-1].leaf = token.piece_designator
        c[-1].setfilter = True
        self.b_pop_nested_stack_to_implied_brace()
        self.cql_tokens.append(self.tokens.pop(0))
        self.b_collapse_parameters_stack()

    def add_left_parenthesis(self, token):
        self.node_stack[-1].children.append(
            self.create_node(token.left_parenthesis))
        self.node_stack.append(self.node_stack[-1].children[-1])
        self.variables_stack.append(set())
        self.cql_tokens.append(self.tokens.pop(0))

    def add_piece_square_variable(self, token):
        psv = token.piece_square_variable
        if psv not in self.variables:
            self.stack.append(self.p_error)
            return True
        self.node_stack[-1].children.append(self.create_node(psv[0]))
        self.node_stack[-1].children[-1].leaf = self.variables[psv]
        self.b_pop_nested_stack()
        self.cql_tokens.append(self.tokens.pop(0))

    def add_plain_filter(self, token):
        c = self.node_stack[-1].children
        c.append(self.create_node(constants.PLAIN_FILTER))
        c[-1].leaf = token.plain
        if token.plain in (constants.DARKSQUARES, constants.LIGHTSQUARES):
            c[-1].setfilter = True
        else:
            c[-1].setfilter = False
        self.b_pop_nested_stack_to_implied_brace()
        self.cql_tokens.append(self.tokens.pop(0))
        self.b_collapse_parameters_stack()

    def add_move_from_to_enpassantsquare_parameter(self, token):
        n = self.create_node(token.move_parameter)
        self.node_stack[-1].children[-1].children.append(n)
        self.cql_tokens.append(self.tokens.pop(0))
        self.node_stack.append(n)
        self.stack.append(self.b_body)
        self.variables_stack.append(set())

    # 'K' '[kb]' 'k or b' 'k on b' are acceptable values for promote parameter
    # but 'a1' 'ka1' 'shift k' are not.
    # Separate method to be done.
    add_move_promote_parameter = add_move_from_to_enpassantsquare_parameter

    def b_move(self):
        if not self.tokens:
            self.b_end()
            return
        token = Token(*self.tokens[0])
        t = token.mainline
        if not t:
            t = token.next_previous
        if not t:
            t = token.move_parameter
        if len(self.node_stack[-1].children):
            if t:
                self.stack.append(self.p_error)
                return True
        elif t in MOVE_MODIFIERS:
            self.node_stack[-1].children.append(self.create_node(t))
            self.cql_tokens.append(self.tokens.pop(0))
        else:
            self.node_stack[-1].children.append(
                self.create_node(constants.NEXT))
        self.stack[-1] = self.b_move_from

    def b_move_from(self):
        if not self.tokens:
            self.b_end()
            return
        token = Token(*self.tokens[0])
        self.stack[-1] = self.b_move_to
        if token.move_parameter == constants.FROM:
            self.add_move_from_to_enpassantsquare_parameter(token)

    def b_move_to(self):
        if not self.tokens:
            self.b_end()
            return
        token = Token(*self.tokens[0])
        self.stack[-1] = self.b_move_promote
        if token.move_parameter == constants.TO:
            self.add_move_from_to_enpassantsquare_parameter(token)

    def b_move_promote(self):
        if not self.tokens:
            self.b_end()
            return
        token = Token(*self.tokens[0])
        self.stack[-1] = self.b_move_enpassantsquare
        if token.move_parameter == constants.PROMOTE:
            self.add_move_promote_parameter(token)

    def b_move_enpassantsquare(self):
        if not self.tokens:
            self.b_end()
            return
        token = Token(*self.tokens[0])
        if token.move_parameter == constants.ENPASSANTSQUARE:
            self.add_move_from_to_enpassantsquare_parameter(token)
            return
        elif token.move_parameter == constants.ENPASSANT:
            n = self.create_node(constants.ENPASSANTSQUARE)
            self.node_stack[-1].children[-1].children.append(n)
            n.leaf = 'a-h1-8'
            self.cql_tokens.append(self.tokens.pop(0))
        self.pop_top_of_stack()

    def b_two_set_filters(self):
        if not self.tokens:
            self.b_end()
            return
        token = Token(*self.tokens[0])
        if token.right_parenthesis:
            tns = self.node_stack[-1]
            if tns.type != constants.LEFT_PARENTHESIS_FILTER:
                self.stack.append(self.p_error)
                return True

            # Exactly two set-filters required.
            elif len(tns.children) != 2:
                self.stack.append(self.p_error)
                return True

            if not tns.set_filter_type_all_children_setfilters():
                self.stack.append(self.p_error)
                return True
            self.b_pop_variables_stack()
            self.b_pop_top_of_stack()
            self.cql_tokens.append(self.tokens.pop(0))

        # Exactly two set-filters required.
        elif len(self.node_stack[-1].children) > 1:
            self.stack.append(self.p_error)
            return True

        else:
            self.b_collapse_filters_stack()
            return self.add_set_body(token, self.b_two_set_filters)

    def b_one_set_filter(self):
        if not self.tokens:
            self.b_end()
            return
        token = Token(*self.tokens[0])

        # Exactly one set-filter required.
        tns = self.node_stack[-1]
        c = len(tns.children)
        if c > 1:
            self.stack.append(self.p_error)
            return True
        elif c:
            if not tns.set_filter_type_all_children_setfilters():
                self.stack.append(self.p_error)
                return True
            self.b_pop_top_of_stack()

        else:
            self.b_collapse_filters_stack()
            return self.add_set_body(token, self.b_one_set_filter)

    def b_two_or_more_set_filters(self):
        if not self.tokens:
            self.b_end()
            return
        token = Token(*self.tokens[0])
        if token.right_parenthesis:
            tns = self.node_stack[-1]
            if tns.type != constants.LEFT_PARENTHESIS_FILTER:
                self.stack.append(self.p_error)
                return True
            elif len(tns.children) < 2:
                self.stack.append(self.p_error)
                return True
            if not tns.set_filter_type_all_children_setfilters():
                self.stack.append(self.p_error)
                return True
            self.b_pop_variables_stack()
            self.b_pop_top_of_stack()
            self.cql_tokens.append(self.tokens.pop(0))
        else:
            self.b_collapse_filters_stack()
            return self.add_set_body(token, self.b_two_or_more_set_filters)

    def b_one_or_more_re_set_filters(self):
        if not self.tokens:
            self.b_end()
            return
        token = Token(*self.tokens[0])
        if token.right_parenthesis:
            tns = self.node_stack[-1]
            if tns.type != constants.LEFT_PARENTHESIS_FILTER:
                self.stack.append(self.p_error)
                return True
            elif len(tns.children) < 1:
                self.stack.append(self.p_error)
                return True
            if not tns.set_filter_type_all_children_setfilters():
                self.stack.append(self.p_error)
                return True
            self.b_pop_variables_stack()
            self.b_pop_top_of_stack()
            self.cql_tokens.append(self.tokens.pop(0))
        else:
            self.b_collapse_filters_stack()
            return self.add_re_cql_body(
                token, self.b_one_or_more_re_set_filters)

    def b_one_or_more_filters(self):
        if not self.tokens:
            self.b_end()
            return
        token = Token(*self.tokens[0])
        if token.right_parenthesis:
            tns = self.node_stack[-1]
            if tns.type != constants.LEFT_PARENTHESIS_FILTER:
                self.stack.append(self.p_error)
                return True
            elif len(tns.children) < 1:
                self.stack.append(self.p_error)
                return True
            if not tns.set_filter_type_all_children_setfilters():
                self.stack.append(self.p_error)
                return True
            self.b_pop_variables_stack()
            self.b_pop_top_of_stack()
            self.cql_tokens.append(self.tokens.pop(0))
        else:
            self.b_collapse_filters_stack()
            return self.add_cql_body(token, self.b_one_or_more_filters)

    def b_attack(self):
        if not self.tokens:
            self.b_end()
            return
        elif len(self.node_stack[-1].children) == 1:
            self.b_pop_top_of_stack()
            return
        token = Token(*self.tokens[0])
        if token.left_parenthesis:
            self.add_left_parenthesis(token)
            self.stack.append(self.b_two_set_filters)
        elif token.number:
            if self.node_stack[-1].range:
                self.stack.append(self.p_error)
                return True
            self.node_stack.append(self.node_stack[-1])
            self.stack.append(self.p_range)
        else:
            self.stack.append(self.p_error)
            return True

    def b_between(self):
        if not self.tokens:
            self.b_end()
            return
        elif len(self.node_stack[-1].children) == 1:
            self.b_pop_top_of_stack()
            return
        token = Token(*self.tokens[0])
        if token.left_parenthesis:
            self.add_left_parenthesis(token)
            self.stack.append(self.b_two_set_filters)
        else:
            self.stack.append(self.p_error)
            return True

    def b_countsquares_power(self):
        if not self.tokens:
            self.b_end()
            return
        elif len(self.node_stack[-1].children) == 1:
            self.b_pop_top_of_stack()
            return
        token = Token(*self.tokens[0])
        if token.number:
            if self.node_stack[-1].range:
                self.stack.append(self.p_error)
                return True
            self.node_stack.append(self.node_stack[-1])
            self.node_stack.append(self.node_stack[-1])
            self.stack.append(self.b_one_set_filter)
            self.stack.append(self.p_range)
        else:
            self.stack.append(self.p_error)
            return True

    def b_powerdifference(self):
        if not self.tokens:
            self.b_end()
            return
        elif len(self.node_stack[-1].children) == 1:
            self.b_pop_top_of_stack()
            return
        token = Token(*self.tokens[0])
        if token.left_parenthesis:
            if not self.node_stack[-1].range:
                self.stack.append(self.p_error)
                return True
            self.add_left_parenthesis(token)
            self.stack.append(self.b_two_set_filters)
        elif token.number:
            if self.node_stack[-1].range:
                self.stack.append(self.p_error)
                return True
            self.node_stack.append(self.node_stack[-1])
            self.stack.append(self.p_range)
        else:
            self.stack.append(self.p_error)
            return True

    def b_ray(self):
        if not self.tokens:
            self.b_end()
            return
        elif len(self.node_stack[-1].children) == 1:
            self.b_pop_top_of_stack()
            return
        token = Token(*self.tokens[0])
        if token.left_parenthesis:
            self.add_left_parenthesis(token)
            self.stack.append(self.b_two_or_more_set_filters)
        elif token.number:
            if self.node_stack[-1].range:
                self.stack.append(self.p_error)
                return True
            self.node_stack.append(self.node_stack[-1])
            self.stack.append(self.p_range)

        # These can be filters by themselves, when their child is never a '(',
        # but are direction keywords for the ray filter here.
        elif token.attack:
            self.add_argument(token.attack, self.b_ray)
        elif token.transform in constants.DIRECTION_FILTER:
            self.add_argument(token.transform, self.b_ray)

        else:
            self.stack.append(self.p_error)
            return True

    def b_next_previous(self):
        if not self.tokens:
            self.b_end()
            return
        elif len(self.node_stack[-1].children) == 1:
            self.b_pop_top_of_stack()
            return
        token = Token(*self.tokens[0])
        if token.left_parenthesis:
            self.add_left_parenthesis(token)
            self.stack.append(self.b_one_or_more_re_set_filters)
        elif token.number:
            if self.node_stack[-1].range:
                self.stack.append(self.p_error)
                return True
            self.node_stack.append(self.node_stack[-1])
            self.stack.append(self.p_range)
        else:
            self.stack.append(self.p_error)
            return True

    def b_relation(self):
        # Copied from http://www.gadycosteff.com/cql/doc/relation.html (cql5.1)
        # relation := 'relation' targetfilters relationparameters
        # targetfilters := filter*
        # relationparameters := relation_parameter+
        #
        # The b_relation_*() methods make the relation syntax equivalent to
        # relation := 'relation' '(' targetfilters relationparameters ')' ...
        # because filter can be {...} and relation_parameter can be (...).
        if not self.tokens:
            self.b_end()
            return
        token = Token(*self.tokens[0])
        t = token.relation_parameter
        if t in RELATION_ECHO_PARAMETERS:
            self.stack[-1] = self.b_relation_parameter
        elif t:
            self.stack.append(self.p_error)
            return True
        elif token.left_parenthesis:
            self.stack[-1] = self.b_relation_parameter
        else:
            return self.add_cql_body(token, self.b_relation)

    def b_relation_parameter(self):
        # Assume only calls are via b_relation().
        if not self.tokens:
            self.b_end()
            return
        token = Token(*self.tokens[0])
        t = token.relation_parameter
        if t in RELATION_ECHO_PARAMETERS:
            self.add_argument(t, self.b_relation_parameter_left_parenthesis)
        elif t:
            self.stack.append(self.p_error)
            return True
        elif token.left_parenthesis:
            self.stack[-1] = self.b_relation_parameter_left_parenthesis
        else:
            # relation is treated as implying parentheses because '(...)' and
            # '{...}' can repeat as part of relation filter.
            self.node_stack[-1].setfilter = False
            self.b_pop_top_of_stack_and_nested()

    def b_relation_parameter_left_parenthesis(self):
        # Assume only calls are via b_relation_parameter().
        if not self.tokens:
            self.b_end()
            return
        token = Token(*self.tokens[0])
        if token.right_parenthesis:
            self.stack[-1] = self.b_relation_parameter
            self.cql_tokens.append(self.tokens.pop(0))
            if not self.tokens:
                if self.node_stack[-1].type == constants.RELATION:
                    self.b_pop_top_of_stack()
            return
        elif not token.left_parenthesis:
            self.stack.append(self.p_error)
            return True
        self.add_argument(token.left_parenthesis,
                          self.b_relation_parameter_type)

    def b_relation_parameter_type(self):
        if not self.tokens:
            self.b_end()
            return
        token = Token(*self.tokens[0])
        if token.relation_parameter in RELATION_LCA_RANGE_PARAMETERS:
            self.add_argument(token.relation_parameter, self.p_range)
            self.node_stack[-1].leaf = self.node_stack[-1].range
        elif token.relation_parameter in RELATION_LCA_PLAIN_PARAMETERS:
            self.add_argument_leaf(constants.RELATION_PARAMETER_NAME,
                                   token.relation_parameter)
        elif token.relation_parameter == constants.TOMOVE:
            self.add_argument(token.relation_parameter,
                              self.b_relation_tomove_argument)
        elif token.relation_parameter in RELATION_SQUARE_PARAMETERS:
            if token.relation_parameter in RELATION_TOMOVE_ARGUMENTS:
                self.add_argument(token.relation_parameter,
                                  self.b_relation_square_tomove_argument)
            else:
                self.add_argument(token.relation_parameter,
                                  self.b_one_set_filter)
        elif token.right_parenthesis:
            self.pop_top_of_stack()
        else:
            self.stack.append(self.p_error)
            return True

    def b_relation_tomove_argument(self):
        if not self.tokens:
            self.b_end()
            return
        token = Token(*self.tokens[0])
        if token.relation_parameter in RELATION_TOMOVE_ARGUMENTS:
            self.node_stack[-1].leaf = token.relation_parameter
            self.cql_tokens.append(self.tokens.pop(0))
            if not self.tokens:
                self.stack.append(self.p_error)
                return True
            if not Token(*self.tokens[0]).right_parenthesis:
                self.stack.append(self.p_error)
                return True
            self.stack[-1] = self.b_relation_parameter_type
        else:
            self.stack.append(self.p_error)
            return True

    def b_relation_square_tomove_argument(self):
        if not self.tokens:
            self.b_end()
            return
        ns = self.node_stack
        if not ns[-1].range:
            ns.append(ns[-1])
            self.stack.append(self.p_numbers)
            return
        ns[-1].leaf = ns[-1].range
        ns[-1].setfilter = False
        self.b_pop_top_of_stack_and_nested()

    def b_elo_white_black(self):
        if not self.tokens:
            self.b_end()
            return
        ns = self.node_stack
        token = Token(*self.tokens[0])
        if token.white_black_keywords:
            if ns[-1].range:
                self.stack.append(self.p_error)
                return True
            self.add_argument(token.white_black_keywords, self.b_numbers)
            ns[-1].leaf = ns[-1].range
        elif token.number:
            if len(ns[-1].children):
                self.stack.append(self.p_error)
                return True
            ns.append(ns[-1])
            self.stack.append(self.b_numbers)
            ns[-1].leaf = ns[-1].range
        elif ns[-1].type == constants.ELO:
            self.pop_top_of_stack()
        else:
            self.stack.append(self.p_error)
            return True

    def b_player_white_black(self):
        if not self.tokens:
            self.b_end()
            return
        ns = self.node_stack
        token = Token(*self.tokens[0])
        if token.white_black_keywords:
            if ns[-1].leaf:
                self.stack.append(self.p_error)
                return True
            self.add_argument(token.white_black_keywords,
                              self.p_double_quoted_string)
        elif token.double_quoted_string:
            if len(ns[-1].children):
                self.stack.append(self.p_error)
                return True
            ns.append(ns[-1])
            self.stack.append(self.p_double_quoted_string)
        elif ns[-1].type == constants.PLAYER:
            self.pop_top_of_stack()
        else:
            self.stack.append(self.p_error)
            return True

    def _is_binary_operator_not_allowed(self):

        # Prior token can indicate with certainty 'or' or 'on' is not allowed,
        # but not that 'or' or 'on' is allowed.  For the latter case assume
        # stack has been collapsed to context expecting a filter.
        ct = self.cql_tokens
        if len(ct):
            if (ct[-1][constants.NOT_INDEX] or
                ct[-1][constants.OR_INDEX] or
                ct[-1][constants.ON_INDEX] or
                ct[-1][constants.LEFT_BRACE_INDEX]
                ):
                self.stack.append(self.p_error)
                return True
        else:
            self.stack.append(self.p_error)
            return True
        if not self.node_stack[-1].children and not self.node_stack[-1].leaf:
            self.stack.append(self.p_error)
            return True

    # Rename as add_argument_child?
    def add_argument(self, token_identity, method):
        ns = self.node_stack
        ns[-1].children.append(self.create_node(token_identity))
        self.node_stack.append(ns[-1].children[-1])
        self.stack.append(method)
        self.cql_tokens.append(self.tokens.pop(0))

    # Add method argument following add_argument style?
    def add_argument_leaf(self, token_identity, leaf):
        ns = self.node_stack
        ns[-1].children.append(self.create_node(token_identity))
        ns[-1].children[-1].leaf = leaf
        self.cql_tokens.append(self.tokens.pop(0))

    def add_binary_argument(self, token_identity, method):

        # Binding is:
        # { check <op> { Ra3 <op> { next ( any Qa2 ) <op> hascomment "eh?" } } }
        if self._is_binary_operator_not_allowed():
            return True
        n = self.create_node(token_identity)
        ns = self.node_stack
        n.children.append(ns[-1].children[-1])
        ns[-1].children[-1] = n
        ns.append(n)
        self.stack.append(method)
        self.cql_tokens.append(self.tokens.pop(0))

    def b_piece_square_filter(self):
        token = Token(*self.tokens[0])
        if token.piece_square_keywords == constants.ALL:
            if self.node_stack[-1].type != constants.SQUARE:
                self.stack.append(self.p_error)
                return True
            self.node_stack[-1].range.extend([token.piece_square_keywords] * 2)
            self.cql_tokens.append(self.tokens.pop(0))
            token = Token(*self.tokens[0])
        if not token.piece_square_variable:
            self.stack.append(self.p_error)
            return True
        if token.piece_square_variable in self.variables:
            self.stack.append(self.p_error)
            return True
        self.node_stack[-1].children.append(
            self.create_node(token.piece_square_variable))
        self.cql_tokens.append(self.tokens.pop(0))
        token = Token(*self.tokens[0])
        if token.piece_square_keywords != constants.IN:
            self.stack.append(self.p_error)
            return True
        self.node_stack.append(self.node_stack[-1].children[-1])
        self.stack.append(self.b_note_piece_square_variable)
        self.cql_tokens.append(self.tokens.pop(0))
        self.node_stack.append(self.create_node(constants.IN))
        self.stack.append(self.b_piece_square_in_set_body)
        self.node_stack[-2].children.append(self.node_stack[-1])

    def b_note_piece_square_variable(self):
        ns = self.node_stack
        if len(ns) < 2:
            self.stack.append(self.p_error)
            return True
        elif ns[-2].type not in VARIABLE_TYPES:
            self.stack.append(self.p_error)
            return True
        self.variables[ns[-1].type] = ns[-2]
        self.variables_stack[-1].add(ns[-1].type)
        self.b_pop_nested_stack()


class ErrorInformation(object):
    """Error information about a ChessQL query and report fomatters.

    This class assumes processing a query stops when the first error is met.

    The CQLStatement parse() and validate() methods return a ErrorInformation
    instance if any attribute other than _statement is bound to an object other
    than None.

    """

    def __init__(self, statement):
        """"""
        self._statement = statement
        self._tokens = None
        self._next_token = None
        self._description = ''

    @property
    def statement(self):
        return self._statement

    @property
    def tokens(self):
        return self._tokens

    @tokens.setter
    def tokens(self, value):
        if self._tokens is not None:
            raise StatementError('A token error already exists.')
        self._tokens = value

    @property
    def next_token(self):
        return self._next_token

    @next_token.setter
    def next_token(self, value):
        if self._next_token is not None:
            raise StatementError('A token error already exists.')
        self._next_token = value

    @property
    def description(self):
        return self._description

    @tokens.setter
    def description(self, value):
        if self._description:
            raise StatementError('A description error already exists.')
        self._description = value

    @property
    def error_found(self):
        return self._tokens is not None or self._description

    def get_error_report(self):
        """Return a str for error dialogue noting token being processed."""
        if self.error_found:
            if self._description:
                return ''.join(
                    ('Error processing statement:\n\n',
                     self._statement.strip(),
                     '\n\n',
                     self._description,
                     '.',
                     ))
            if self._next_token:
                next_token = ''.join(
                    ('\n\nThe item being processed is:\n\n',
                     ''.join(self._next_token),
                     ))
            else:
                next_token = ''
            return ''.join(
                ('Error, likely at or just after end of:\n\n',
                 ' '.join(''.join(t) for t in self._tokens),
                 '\n\nderived from:\n\n',
                 self._statement.strip(),
                 next_token,
                 ))
        return ''.join(('An unidentified error seems to exist in:\n\n',
                         self._statement.strip(),
                         ))

    def add_error_report_to_message(self, message, sep='\n\n'):
        """Return message extended with an error report."""
        return ''.join((message, sep, self.get_error_report()))
    

if __name__ == '__main__':


    # Interactive testing.

    import sys

    # Arguments are the stack[1] entries to cause tracing.
    # This entry will be either p_cql or b_cql at time of writing, referring
    # to parameters or body in 'cql ( <parameters> ) <body>'.

    _trace = frozenset(sys.argv[1:])
    del sys


    class Statement(Statement):

        def parse(self):
            if self._error_information:
                return
            self.stack.append(self.look_for_cql_parameters_or_body)
            while len(self.tokens):
                if self._trace_and_call_top_of_stack():
                    self._trace_and_call_top_of_stack()
                    break
            else:
                self._trace_and_call_top_of_stack()

        def _trace_and_call_top_of_stack(self):
            if len(self.stack) > 1:
                if self.stack[1].__name__ in _trace:
                    print(self._trace(self.stack[-1].__name__))
            return self.stack[-1]()

        def _trace_start(self, method_name):
            return ' '.join(('--', str(method_name)))

        def _trace_body(self):
            ts = [' '.join(('  ', str(len(self.variables)), 'variables'))]
            if self.variables:
                for s in sorted(self.variables):
                    ts.append(' '.join(('     ', s, str(self.variables[s]))))
            ts.append(' '.join(('  ',
                                str(len(self.variables_stack)),
                                'variables_stack')))
            for s in self.variables_stack:
                ts.append(' '.join(('     ', str(s))))
            ts.append(' '.join(('  ', str(len(self.stack)), 'stack')))
            for s in self.stack:
                try:
                    s = s.__name__
                except:
                    pass
                ts.append(' '.join(('     ', s)))
            ts.append(' '.join(('  ', str(len(self.node_stack)), 'node_stack')))
            for s in self.node_stack:
                ts.append(' '.join(('     ', str(s))))
            return '\n'.join(ts)

        def _trace_end(self):
            return '-- end'

        def _trace(self, method_name):
            return '\n'.join((self._trace_start(method_name),
                              self._trace_body(),
                              self._trace_end(),
                              ))


    token_map = {e:v for e, v in enumerate(constants.TOKEN_NAMES.split())}
    while True:
        try:
            s = input('cql> ')
        except EOFError as exc:
            print(exc)
            break
        except KeyboardInterrupt:
            print()
            break
        #if not s:
        #    continue
        try:
            cql = Statement()
            cql.process_statement(s)
            print()
            for t in cql.tokens:
                print(*[(v, token_map[e]) for e, v in enumerate(t) if v][0],
                      sep='\t')
            print(cql.cql_parameters)
            print(cql.cql_filters)
            print(cql.node_stack)
            print(cql.variables_stack)
            if cql.cql_error:
                print('error:', s)
            else:
                print('ok:', s)
            print()
        except RuntimeError as exc:
            print(exc)



# node.py
# Copyright 2017 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Chess Query Language (CQL) statement filter node."""

import copy


class Node:
    """"""
    # self.type == 'not' means the result of the first child node is inverted.
    # 'cql() not p k' produces a 'not' node with two children, 'p' node first.
    # 'cql() not p or k produces a 'not' node with one child, an 'or' node
    #  which has two children, nodes for 'p' and 'q'.
    
    def __init__(self, type_, children=None, leaf=None):
        """"""
        super().__init__()
        self.type = type_
        if children:
            self.children = children
        else:
            self.children = []
        self.leaf = leaf
        self.range = []
        self.repeat = False
        self.setfilter = None

    def __deepcopy__(self, memo):
        """"""
        newcopy = empty_copy(self)
        newcopy.type = self.type
        newcopy.children = copy.deepcopy(self.children, memo)
        newcopy.leaf = copy.deepcopy(self.leaf, memo)
        newcopy.range = copy.deepcopy(self.range, memo)
        newcopy.repeat = self.repeat
        newcopy.setfilter = self.setfilter
        return newcopy

    def __str__(self):
        if not self.range:
            s = str(self.type)
        elif len(self.range) == 1:
            b, = self.range
            s = ''.join(str(i) for i in (self.type, '<', b, ',', b, '>'))
        elif len(self.range) == 2:
            b, e = self.range
            s = ''.join(str(i) for i in (self.type, '<', b, ',', e, '>'))
        else:
            s = ''.join(str(i) for i in (self.type, '<', '?', ',', '?', '>'))
        if self.repeat:
            s = ''.join((s, '<', self.repeat, '>'))
        return ', '.join((str(s),
                          ', '.join((str(c)
                                     for c in self.children)).join(('[', ']')),
                          str(self.leaf),
                          str(self.setfilter),
                          )).join(('(', ')'))

    @property
    def occupied(self):
        return bool(self.children or self.leaf)

    @property
    def children_setfilter_types(self):
        return {n.setfilter for n in self.children}

    def derive_filter_type_from_children(self):
        csft = self.children_setfilter_types
        if len(csft) > 1:
            self.setfilter = False
        elif self.range:
            self.setfilter = False
        else:
            self.setfilter = csft.pop()

    def derive_brace_filter_type_from_children(self):
        if len(self.children) > 1:
            self.setfilter = False
        elif self.range:
            self.setfilter = False
        else:
            self.derive_filter_type_from_children()

    def set_filter_type_all_children_setfilters(self):
        csft = self.children_setfilter_types
        if False in csft or True not in csft:
            return False
        elif self.range:
            self.setfilter = False
            return True
        self.setfilter = True
        return True

    def repeated_parameter(self, t=None):
        """"""
        if t is None:
            t = self.type
        return t in {n.type for n in self.children}

    def get_repeatable_parameter(self, t, node_factory):
        """"""
        for n in self.children:
            if t == n.type:
                return n
        self.children.append(node_factory(t))
        return self.children[-1]


def empty_copy(obj):
    class Empty(obj.__class__):
        def __init__(self):
            pass
    newcopy = Empty()
    newcopy.__class__ = obj.__class__
    return newcopy



# constants.py
# Copyright 2017 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""constants tests for cql"""

import unittest
from copy import copy, deepcopy
import re

from .. import constants


class CQLConstants(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def expected_matches(self, match_spec):
        """"""
        em = []
        for t, p in match_spec:
            m = [''] * 44
            m[p] = t
            em.append(tuple(m))
        return em

    def test__raises(self):
        """"""
        pass

    def test____assumptions(self):
        """"""
        self.assertEqual(constants.FILE_RANGE, r'[a-h]-[a-h]')
        self.assertEqual(constants.RANK_RANGE, r'[1-8]-[1-8]')
        self.assertEqual(constants.RANGE_SEPARATOR, r'-')
        self.assertEqual(constants.COMPOUND_DESIGNATOR_START, r'\[')
        self.assertEqual(constants.COMPOUND_DESIGNATOR_END, r'\]')
        self.assertEqual(constants.FILE_DESIGNATOR, '[a-h]')
        self.assertEqual(constants.RANK_DESIGNATOR, '[1-8]')
        self.assertEqual(constants.SQUARE_DESIGNATOR_SEPARATOR, r',')
        self.assertEqual(constants.FILE_RANGE, '[a-h]-[a-h]')
        self.assertEqual(constants.RANK_RANGE, '[1-8]-[1-8]')
        self.assertEqual(
            constants.SIMPLE_SQUARE_DESIGNATOR,
            ''.join(('(?:[a-h]-[a-h][1-8]-[1-8])|(?:[a-h]-[a-h][1-8])|',
                     '(?:[a-h][1-8]-[1-8])|(?:[a-h][1-8])')))
        self.assertEqual(
            constants.SQUARE_DESIGNATOR,
            ''.join(('(?:(?:[a-h]-[a-h][1-8]-[1-8])|(?:[a-h]-[a-h][1-8])|',
                     '(?:[a-h][1-8]-[1-8])|(?:[a-h][1-8]))|',
                     '(?:\\[(?:(?:[a-h]-[a-h][1-8]-[1-8])|',
                     '(?:[a-h]-[a-h][1-8])|(?:[a-h][1-8]-[1-8])|',
                     '(?:[a-h][1-8]))(?:,(?:(?:[a-h]-[a-h][1-8]-[1-8])|',
                     '(?:[a-h]-[a-h][1-8])|(?:[a-h][1-8]-[1-8])|',
                     '(?:[a-h][1-8])))*\\])')))
        self.assertEqual(constants.WHITE_PIECE_NAMES, r'KQRBNP')
        self.assertEqual(constants.ANY_WHITE_PIECE_NAME, r'A')
        self.assertEqual(constants.BLACK_PIECE_NAMES, r'kqrbnp')
        self.assertEqual(constants.ANY_BLACK_PIECE_NAME, r'a')
        self.assertEqual(constants.EMPTY_SQUARE_NAME, r'.')
        self.assertEqual(constants.PIECE_NAMES, r'KQRBNPkqrbnpAa\.')
        self.assertEqual(
            constants.PIECE_TYPE_DESIGNATOR,
            '(?:[KQRBNPkqrbnpAa\\.])|(?:\\[[KQRBNPkqrbnpAa\\.]+\\])')
        self.assertEqual(
            constants.PIECE_DESIGNATOR,
            ''.join(('(?:(?:(?:(?:[KQRBNPkqrbnpAa\\.])|',
                     '(?:\\[[KQRBNPkqrbnpAa\\.]+\\]))',
                     '(?:(?:(?:[a-h]-[a-h][1-8]-[1-8])|',
                     '(?:[a-h]-[a-h][1-8])|(?:[a-h][1-8]-[1-8])|',
                     '(?:[a-h][1-8]))|(?:\\[(?:(?:[a-h]-[a-h][1-8]-[1-8])|',
                     '(?:[a-h]-[a-h][1-8])|(?:[a-h][1-8]-[1-8])|',
                     '(?:[a-h][1-8]))(?:,(?:(?:[a-h]-[a-h][1-8]-[1-8])|',
                     '(?:[a-h]-[a-h][1-8])|(?:[a-h][1-8]-[1-8])|',
                     '(?:[a-h][1-8])))*\\])))|',
                     '(?:(?:(?:[a-h]-[a-h][1-8]-[1-8])|(?:[a-h]-[a-h][1-8])|',
                     '(?:[a-h][1-8]-[1-8])|(?:[a-h][1-8]))|',
                     '(?:\\[(?:(?:[a-h]-[a-h][1-8]-[1-8])|',
                     '(?:[a-h]-[a-h][1-8])|(?:[a-h][1-8]-[1-8])|',
                     '(?:[a-h][1-8]))(?:,(?:(?:[a-h]-[a-h][1-8]-[1-8])|',
                     '(?:[a-h]-[a-h][1-8])|(?:[a-h][1-8]-[1-8])|',
                     '(?:[a-h][1-8])))*\\]))|(?:(?:[KQRBNPkqrbnpAa\\.])|',
                     '(?:\\[[KQRBNPkqrbnpAa\\.]+\\])))')))
        self.assertEqual(constants.ANY, r'any')
        self.assertEqual(constants.ATTACK, r'attack')
        self.assertEqual(constants.BEGINVARIATION, r'beginvariation')
        self.assertEqual(constants.BETWEEN, r'between')
        self.assertEqual(constants.BTM, r'btm')
        self.assertEqual(constants.CHECK, r'check')
        self.assertEqual(constants.COMMENT, r'comment')
        self.assertEqual(constants.COUNTSQUARES, r'countsquares')
        self.assertEqual(constants.CQL, r'cql')
        self.assertEqual(constants.DARKSQUARES, r'darksquares')
        self.assertEqual(constants.UP, r'up')
        self.assertEqual(constants.DOWN, r'down')
        self.assertEqual(constants.LEFT, r'left')
        self.assertEqual(constants.RIGHT, r'right')
        self.assertEqual(constants.NORTHEAST, r'northeast')
        self.assertEqual(constants.NORTHWEST, r'northwest')
        self.assertEqual(constants.SOUTHEAST, r'southeast')
        self.assertEqual(constants.SOUTHWEST, r'southwest')
        self.assertEqual(constants.VERTICAL, r'vertical')
        self.assertEqual(constants.HORIZONTAL, r'horizontal')
        self.assertEqual(constants.DIAGONAL, r'diagonal')
        self.assertEqual(constants.ORTHOGONAL, r'orthogonal')
        self.assertEqual(constants.ANYDIRECTION, r'anydirection')
        self.assertEqual(constants.ELO, r'elo')
        self.assertEqual(constants.EVENT, r'event')
        self.assertEqual(constants.FLIP, r'flip')
        self.assertEqual(constants.FLIPDIHEDRAL, r'flipdihedral')
        self.assertEqual(constants.FLIPHORIZONTAL, r'fliphorizontal')
        self.assertEqual(constants.FLIPVERTICAL, r'flipvertical')
        self.assertEqual(constants.FLIPCOLOR, r'flipcolor')
        self.assertEqual(constants.GAMENUMBER, r'gamenumber')
        self.assertEqual(constants.HASCOMMENT, r'hascomment')
        self.assertEqual(constants.INITIAL, r'initial')
        self.assertEqual(constants.INPUT, r'input')
        self.assertEqual(constants.LIGHTSQUARES, r'lightsquares')
        self.assertEqual(constants.MAINLINE, r'mainline')
        self.assertEqual(constants.MATCHCOUNT, r'matchcount')
        self.assertEqual(constants.MATE, r'mate')
        self.assertEqual(constants.MOVE, r'move')
        self.assertEqual(constants.FROM, r'from')
        self.assertEqual(constants.TO, r'to')
        self.assertEqual(constants.PROMOTE, r'promote')
        self.assertEqual(constants.MAINLINE, r'mainline')
        self.assertEqual(constants.PREVIOUS, r'previous')
        self.assertEqual(constants.ENPASSANT, r'enpassant')
        self.assertEqual(constants.ENPASSANTSQUARE, r'enpassantsquare')
        self.assertEqual(constants.MOVENUMBER, r'movenumber')
        self.assertEqual(constants.NEXT, r'next')
        self.assertEqual(constants.NEXT2, r'next2')
        self.assertEqual(constants.NEXT_STAR, r'next\*')
        self.assertEqual(constants.NOT, r'not')
        self.assertEqual(constants.ON, r'on')
        self.assertEqual(constants.OR, r'or')
        self.assertEqual(constants.ORIGIN, r'origin')
        self.assertEqual(constants.OUTPUT, r'output')
        self.assertEqual(constants.PIECE, r'piece')
        self.assertEqual(constants.IN, r'in')
        self.assertEqual(constants.PLAYER, r'player')
        self.assertEqual(constants.WHITE, r'white')
        self.assertEqual(constants.BLACK, r'black')
        self.assertEqual(constants.POWER, r'power')
        self.assertEqual(constants.POWERDIFFERENCE, r'powerdifference')
        self.assertEqual(constants.PREVIOUS, r'previous')
        self.assertEqual(constants.PREVIOUS2, r'previous2')
        self.assertEqual(constants.PREVIOUS_STAR, r'previous\*')
        self.assertEqual(constants.RAY, r'ray')
        self.assertEqual(constants.RELATION, r'relation')
        self.assertEqual(constants.ANCESTOR, r'ancestor')
        self.assertEqual(constants.DESCENDANT, r'descendant')
        self.assertEqual(constants.ECHOFLIP, r'echoflip')
        self.assertEqual(constants.ECHOFLIPVERTICAL, r'echoflipvertical')
        self.assertEqual(constants.ECHOFLIPHORIZONTAL, r'echofliphorizontal')
        self.assertEqual(constants.ECHOSHIFT, r'echoshift')
        self.assertEqual(constants.ECHOSHIFTHORIZONTAL, r'echoshifthorizontal')
        self.assertEqual(constants.ECHOSHIFTVERTICAL, r'echoshiftvertical')
        self.assertEqual(constants.ECHOROTATE45, r'echorotate45')
        self.assertEqual(constants.ECHOROTATE90, r'echorotate90')
        self.assertEqual(constants.LCAMAX, r'lcamax')
        self.assertEqual(constants.LCASOURCE, r'lcasource')
        self.assertEqual(constants.LCASUM, r'lcasum')
        self.assertEqual(constants.LCATARGET, r'lcatarget')
        self.assertEqual(constants.LCASUBSTRING, r'lcasubstring')
        self.assertEqual(constants.MATCH, r'match')
        self.assertEqual(constants.MISMATCH, r'mismatch')
        self.assertEqual(constants.SOURCESQUARES, r'sourcesquares')
        self.assertEqual(constants.TARGETSQUARES, r'targetsquares')
        self.assertEqual(constants.TOMOVE, r'tomove')
        self.assertEqual(constants.RESULT, r'result')
        self.assertEqual(constants.ROTATE45, r'rotate45')
        self.assertEqual(constants.ROTATE90, r'rotate90')
        self.assertEqual(constants.SHIFT, r'shift')
        self.assertEqual(constants.SHIFTHORIZONTAL, r'shifthorizontal')
        self.assertEqual(constants.SHIFTVERTICAL, r'shiftvertical')
        self.assertEqual(constants.SILENT, r'silent')
        self.assertEqual(constants.SITE, r'site')
        self.assertEqual(constants.SORT, r'sort')
        self.assertEqual(constants.SQUARE, r'square')
        self.assertEqual(constants.ALL, r'all')
        self.assertEqual(constants.IN, r'in')
        self.assertEqual(constants.STALEMATE, r'stalemate')
        self.assertEqual(constants.TERMINAL, r'terminal')
        self.assertEqual(constants.VARIATION, r'variation')
        self.assertEqual(constants.VARIATIONS, r'variations')
        self.assertEqual(constants.WTM, r'wtm')
        self.assertEqual(constants.YEAR, r'year')
        self.assertEqual(constants.PIECE_SQUARE, r'square|piece')
        self.assertEqual(constants.PIECE_SQUARE_KEYWORDS, r'all|in')
        self.assertEqual(constants.WHITE_BLACK_KEYWORDS, r'white|black')
        self.assertEqual(constants.SITE_EVENT, r'event|site')
        self.assertEqual(constants.COUNTSQUARES_POWER, r'countsquares|power')
        self.assertEqual(constants.NEXT_PREVIOUS,
                         r'previous2|previous|next2|next')
        self.assertEqual(constants.PLAIN,
                         r'|'.join((r'beginvariation',
                                    r'lightsquares',
                                    r'darksquares',
                                    r'variation',
                                    r'stalemate',
                                    r'terminal',
                                    r'initial',
                                    r'check',
                                    r'mate',
                                    r'wtm',
                                    r'btm',
                                    r'any',
                                    )))
        self.assertEqual(constants.PLAIN_FILTER, 'plain')
        self.assertEqual(constants.RELATION_PARAMETER,
                         r'|'.join((r'echoshifthorizontal',
                                    r'echofliphorizontal',
                                    r'echoshiftvertical',
                                    r'echoflipvertical',
                                    r'targetsquares',
                                    r'sourcesquares',
                                    r'lcasubstring',
                                    r'echorotate90',
                                    r'echorotate45',
                                    r'descendant',
                                    r'lcatarget',
                                    r'lcasource',
                                    r'echoshift',
                                    r'mismatch',
                                    r'echoflip',
                                    r'ancestor',
                                    r'tomove',
                                    r'lcasum',
                                    r'lcamax',
                                    r'match',
                                    )))
        self.assertEqual(constants.MOVE_PARAMETER,
                         r'enpassantsquare|enpassant|promote|empty|from|to')
        self.assertEqual(constants.CQL_PARAMETER,
                         r'variations|matchcount|gamenumber|output|input')
        self.assertEqual(constants.CYCLIC_FILTER,
                         {r'up',
                          r'down',
                          r'left',
                          r'right',
                          r'northeast',
                          r'northwest',
                          r'southeast',
                          r'southwest',
                          })
        self.assertEqual(constants.DIRECTION_FILTER,
                         {r'up',
                          r'down',
                          r'left',
                          r'right',
                          r'northeast',
                          r'northwest',
                          r'southeast',
                          r'southwest',
                          r'vertical',
                          r'horizontal',
                          r'diagonal',
                          r'orthogonal',
                          r'anydirection',
                          })
        self.assertEqual(constants.DIHEDRAL_FILTER,
                         {r'flip',
                          r'flipdihedral',
                          r'fliphorizontal',
                          r'flipvertical',
                          r'rotate90',
                          })
        self.assertEqual(constants.SHIFT_FILTER,
                         {r'shifthorizontal',
                          r'shiftvertical',
                          r'shift',
                          })
        self.assertEqual(constants.TRANSFORM,
                         r'|'.join((r'shifthorizontal',
                                    r'fliphorizontal',
                                    r'shiftvertical',
                                    r'flipvertical',
                                    r'flipdihedral',
                                    r'anydirection',
                                    r'previous\*',
                                    r'orthogonal',
                                    r'horizontal',
                                    r'southwest',
                                    r'southeast',
                                    r'northwest',
                                    r'northeast',
                                    r'flipcolor',
                                    r'vertical',
                                    r'rotate90',
                                    r'rotate45',
                                    r'diagonal',
                                    r'next\*',
                                    r'shift',
                                    r'right',
                                    r'left',
                                    r'flip',
                                    r'down',
                                    r'up',
                                    )))
        self.assertEqual(constants.LEFT_PARENTHESIS, r'\(')
        self.assertEqual(constants.RIGHT_PARENTHESIS, r'\)')
        self.assertEqual(constants.LEFT_BRACE, r'\{')
        self.assertEqual(constants.RIGHT_BRACE, r'\}')
        self.assertEqual(constants.NUMBER, r'\d+')
        self.assertEqual(constants.DOUBLE_QUOTED_STRING,
                         r'"[^\\"]*(?:\\.[^\\"]*)*"')
        self.assertEqual(constants.COMMENT_STRING, r';[^\n]*')
        self.assertEqual(constants.PGN_FILE, r'\.pgn')
        self.assertEqual(constants.ALLOWED_STRINGS,
                         r'(?:1-0)|(?:0-1)|(?:1/2-1/2)|(?:[^\s]+\.pgn)')
        self.assertEqual(constants.ZERO_OR_MORE, r'\*')
        self.assertEqual(constants.ONE_OR_MORE, r'\+')
        self.assertEqual(constants.REPEAT_OPERATORS, r'\?|\*|\+')
        self.assertEqual(constants.BAD_TOKEN, r'\S+')
        self.assertEqual(constants.EDGING, r'\s|\(|\)|\{|\}')
        self.assertEqual(constants.REPEAT_EDGING, r'\s|\(|\)|\{|\}|\?|\*|\+')
        self.assertEqual(constants.LEADING, '(?<=\\s|\\(|\\)|\\{|\\})')
        self.assertEqual(constants.TRAILING, '(?=\s|\(|\)|\{|\}|\Z)')
        self.assertEqual(constants.REPEAT_TRAILING,
                         '(?=\s|\(|\)|\{|\}|\?|\*|\+|\Z)')
        self.assertEqual(
            constants.CQL_PATTERN,
            ''.join((
                '(\\()|',
                '(\\))|',
                '(\\{)|',
                '(\\})|',
                '(;[^\\n]*)|',
                '(?:\\A\\s*)(cql)(?=\\s|',
                '\\()|',
                '(?<=\\s|',
                '\\(|',
                '\\)|',
                '\\{|',
                '\\})((?:(?:(?:(?:[KQRBNPkqrbnpAa\\.])|',
                '(?:\\[[KQRBNPkqrbnpAa\\.]+\\]))(?:(?:(?:[a-h]-[a-h][1-8]-[1-8])|',
                '(?:[a-h]-[a-h][1-8])|',
                '(?:[a-h][1-8]-[1-8])|',
                '(?:[a-h][1-8]))|',
                '(?:\\[(?:(?:[a-h]-[a-h][1-8]-[1-8])|',
                '(?:[a-h]-[a-h][1-8])|',
                '(?:[a-h][1-8]-[1-8])|',
                '(?:[a-h][1-8]))(?:,(?:(?:[a-h]-[a-h][1-8]-[1-8])|',
                '(?:[a-h]-[a-h][1-8])|',
                '(?:[a-h][1-8]-[1-8])|',
                '(?:[a-h][1-8])))*\\])))|',
                '(?:(?:(?:[a-h]-[a-h][1-8]-[1-8])|',
                '(?:[a-h]-[a-h][1-8])|',
                '(?:[a-h][1-8]-[1-8])|',
                '(?:[a-h][1-8]))|',
                '(?:\\[(?:(?:[a-h]-[a-h][1-8]-[1-8])|',
                '(?:[a-h]-[a-h][1-8])|',
                '(?:[a-h][1-8]-[1-8])|',
                '(?:[a-h][1-8]))(?:,(?:(?:[a-h]-[a-h][1-8]-[1-8])|',
                '(?:[a-h]-[a-h][1-8])|',
                '(?:[a-h][1-8]-[1-8])|',
                '(?:[a-h][1-8])))*\\]))|',
                '(?:(?:[KQRBNPkqrbnpAa\\.])|',
                '(?:\\[[KQRBNPkqrbnpAa\\.]+\\]))))(?=\\s|',
                '\\(|\\)|\\{|\\}|\\?|\\*|\\+|\\Z)|',
                '(?<=\\s|\\(|\\)|\\{|\\})(variations|',
                'matchcount|gamenumber|output|',
                'input)(?=\\s|\\(|\\)|\\{|\\}|\\Z)|',
                '(?<=\\s|\\(|\\)|\\{|\\})(shifthorizontal|',
                'fliphorizontal|shiftvertical|flipvertical|flipdihedral|',
                'anydirection|previous\\*|orthogonal|horizontal|',
                'southwest|southeast|northwest|northeast|flipcolor|',
                'vertical|rotate90|rotate45|diagonal|next\\*|shift|',
                'right|left|flip|down|up)(?=\\s|\\(|\\)|\\{|\\}|\\Z)|',
                '(?<=\\s|\\(|\\)|\\{|\\})(beginvariation|',
                'lightsquares|darksquares|variation|stalemate|terminal|',
                'initial|check|mate|wtm|btm|',
                'any)(?=\\s|\\(|\\)|\\{|\\}|\\?|\\*|\\+|\\Z)|',
                '(?<=\\s|\\(|\\)|\\{|\\})(echoshifthorizontal|',
                'echofliphorizontal|echoshiftvertical|echoflipvertical|',
                'targetsquares|sourcesquares|lcasubstring|echorotate90|',
                'echorotate45|descendant|lcatarget|lcasource|echoshift|',
                'mismatch|echoflip|ancestor|tomove|lcasum|lcamax|',
                'match)(?=\\s|\\(|\\)|\\{|\\}|\\Z)|',
                '(?<=\\s|\\(|\\)|\\{|\\})(enpassantsquare|',
                'enpassant|promote|empty|from|',
                'to)(?=\\s|\\(|\\)|\\{|\\}|\\Z)|',
                '(?<=\\s|\\(|\\)|\\{|\\})(\\d+)(?=\\s|\\(|\\)|\\{|\\}|\\?|\\*|\\+|\\Z)|',
                '("[^\\\\"]*(?:\\\\.[^\\\\"]*)*")|',
                '(?<=\\s|\\(|\\)|\\{|\\})(or)(?=\\s|\\(|\\)|\\{|\\}|\\Z)|',
                '(?<=\\s|\\(|\\)|\\{|\\})(not)(?=\\s|\\(|\\)|\\{|\\}|\\Z)|',
                '(?<=\\s|\\(|\\)|\\{|\\})(attack)(?=\\s|\\(|\\)|\\{|\\}|\\Z)|',
                '(?<=\\s|\\(|\\)|\\{|\\})(between)(?=\\s|\\(|\\)|\\{|\\}|\\Z)|',
                '(?<=\\s|\\(|\\)|\\{|\\})(countsquares|',
                'power)(?=\\s|\\(|\\)|\\{|\\}|\\Z)|',
                '(?<=\\s|\\(|\\)|\\{|\\})(elo)(?=\\s|\\(|\\)|\\{|\\}|\\Z)|',
                '(?<=\\s|\\(|\\)|\\{|\\})(comment)(?=\\s|\\(|\\)|\\{|\\}|\\Z)|',
                '(?<=\\s|\\(|\\)|\\{|\\})(hascomment)(?=\\s|\\(|\\)|\\{|\\}|\\Z)|',
                '(?<=\\s|\\(|\\)|\\{|\\})(mainline)(?=\\s|\\(|\\)|\\{|\\}|\\Z)|',
                '(?<=\\s|\\(|\\)|\\{|\\})(move)(?=\\s|\\(|\\)|\\{|\\}|\\Z)|',
                '(?<=\\s|\\(|\\)|\\{|\\})(movenumber)(?=\\s|\\(|\\)|\\{|\\}|\\Z)|',
                '(?<=\\s|\\(|\\)|\\{|\\})(previous2|',
                'previous|next2|next)(?=\\s|\\(|\\)|\\{|\\}|\\Z)|',
                '(?<=\\s|\\(|\\)|\\{|\\})(on)(?=\\s|\\(|\\)|\\{|\\}|\\Z)|',
                '(?<=\\s|\\(|\\)|\\{|\\})(player)(?=\\s|\\(|\\)|\\{|\\}|\\Z)|',
                '(?<=\\s|\\(|\\)|\\{|\\})(powerdifference)(?=\\s|\\(|\\)|\\{|\\}|\\Z)|',
                '(?<=\\s|\\(|\\)|\\{|\\})(ray)(?=\\s|\\(|\\)|\\{|\\}|\\Z)|',
                '(?<=\\s|\\(|\\)|\\{|\\})(relation)(?=\\s|\\(|\\)|\\{|\\}|\\Z)|',
                '(?<=\\s|\\(|\\)|\\{|\\})(result)(?=\\s|\\(|\\)|\\{|\\}|\\Z)|',
                '(?<=\\s|\\(|\\)|\\{|\\})(sort)(?=\\s|\\(|\\)|\\{|\\}|\\Z)|',
                '(?<=\\s|\\(|\\)|\\{|\\})(year)(?=\\s|\\(|\\)|\\{|\\}|\\Z)|',
                '(?<=\\s|\\(|\\)|\\{|\\})(silent)(?=\\s|\\(|\\)|\\{|\\}|\\Z)|',
                '(?<=\\s|\\(|\\)|\\{|\\})(event|site)(?=\\s|\\(|\\)|\\{|\\}|\\Z)|',
                '(?<=\\s|\\(|\\)|\\{|\\})(square|piece)(?=\\s|\\(|\\)|\\{|\\}|\\Z)|',
                '(?<=\\s|\\(|\\)|\\{|\\})(all|in)(?=\\s|\\(|\\)|\\{|\\}|\\Z)|',
                '(?<=\\s|\\(|\\)|\\{|\\})(origin)(?=\\s|\\(|\\)|\\{|\\}|\\Z)|',
                '(?<=\\s|\\(|\\)|\\{|\\})(\\$[a-zA-Z0-9]+)(?=\\s|\\(|\\)|\\{|\\}|\\?|\\*|\\+|\\Z)|',
                '(?<=\\s|\\(|\\)|\\{|\\})(white|black)(?=\\s|\\(|\\)|\\{|\\}|\\Z)|',
                '((?:1-0)|(?:0-1)|(?:1/2-1/2)|(?:[^\\s]+\\.pgn))|(\\?|\\*|\\+)|',
                '(\\S+)',
                )))
        self.assertEqual(constants.TOKEN_NAMES,
                         ' '.join(('left_parenthesis',
                                   'right_parenthesis',
                                   'left_brace',
                                   'right_brace',
                                   'comment_string',
                                   'cql',
                                   'piece_designator',
                                   'cql_parameter',
                                   'transform',
                                   'plain',
                                   'relation_parameter',
                                   'move_parameter',
                                   'number',
                                   'double_quoted_string',
                                   'or_',
                                   'not_',
                                   'attack',
                                   'between',
                                   'countsquares_power',
                                   'elo',
                                   'comment',
                                   'hascomment',
                                   'mainline',
                                   'move',
                                   'movenumber',
                                   'next_previous',
                                   'on',
                                   'player',
                                   'powerdifference',
                                   'ray',
                                   'relation',
                                   'result',
                                   'sort',
                                   'year',
                                   'silent',
                                   'site_event',
                                   'piece_square',
                                   'piece_square_keywords',
                                   'origin',
                                   'piece_square_variable',
                                   'white_black_keywords',
                                   'allowed_strings',
                                   'repeat_operators',
                                   'bad_token',
                                   )))
        self.assertEqual(constants.RANGE, 'range')
        self.assertEqual(constants.NAME_DELIMITER, '\n')
        self.assertEqual(len([c for c in dir(constants)
                              if not c.startswith('__')]), 185)
        self.assertEqual(len(constants.TOKEN_NAMES.split(' ')),
                         len(self.expected_matches((('', 1),))[0]))
        self.assertEqual(constants.RIGHT_PARENTHESIS_INDEX, 1)
        self.assertEqual(constants.LEFT_BRACE_INDEX, 2)
        self.assertEqual(constants.RIGHT_BRACE_INDEX, 3)
        self.assertEqual(constants.PIECE_DESIGNATOR_INDEX, 6)
        self.assertEqual(constants.PLAIN_INDEX, 9)
        self.assertEqual(constants.OR_INDEX, 14)
        self.assertEqual(constants.NOT_INDEX, 15)
        self.assertEqual(constants.ON_INDEX, 26)

    def test_compiled_CQL_PATTERN(self):
        """"""
        pm = re.compile(constants.CQL_PATTERN, flags=re.DOTALL)

        # Examples of each element in the match.
        self.assertEqual(
            pm.findall('('),
            self.expected_matches((('(', 0),)))
        self.assertEqual(
            pm.findall(')'),
            self.expected_matches(((')', 1),)))
        self.assertEqual(
            pm.findall('{'),
            self.expected_matches((('{', 2),)))
        self.assertEqual(
            pm.findall('}'),
            self.expected_matches((('}', 3),)))
        self.assertEqual(
            pm.findall(';comment'),
            self.expected_matches(((';comment', 4),)))
        self.assertEqual(
            pm.findall('cql('),
            self.expected_matches((('cql', 5),('(', 0),)))
        self.assertEqual(
            pm.findall(' cql('),
            self.expected_matches((('cql', 5),('(', 0),)))
        self.assertEqual(
            pm.findall('cql ('),
            self.expected_matches((('cql', 5),('(', 0),)))
        self.assertEqual(
            pm.findall(' cql ('),
            self.expected_matches((('cql', 5),('(', 0),)))
        self.assertEqual(
            pm.findall(' K'),
            self.expected_matches((('K', 6),)))
        self.assertEqual(
            pm.findall('(K'),
            self.expected_matches((('(', 0),('K', 6),)))
        self.assertEqual(
            pm.findall('{K'),
            self.expected_matches((('{', 2),('K', 6),)))
        self.assertEqual(
            pm.findall('( K'),
            self.expected_matches((('(', 0),('K', 6),)))
        self.assertEqual(
            pm.findall('{ K'),
            self.expected_matches((('{', 2),('K', 6),)))
        self.assertEqual(
            pm.findall(' K)'),
            self.expected_matches((('K', 6),(')', 1),)))
        self.assertEqual(
            pm.findall(' K )'),
            self.expected_matches((('K', 6),(')', 1),)))
        self.assertEqual(
            pm.findall(' K}'),
            self.expected_matches((('K', 6),('}', 3),)))
        self.assertEqual(
            pm.findall(' K }'),
            self.expected_matches((('K', 6),('}', 3),)))
        self.assertEqual(
            pm.findall(')K'),
            self.expected_matches(((')', 1),('K', 6),)))
        self.assertEqual(
            pm.findall('}K'),
            self.expected_matches((('}', 3),('K', 6),)))
        self.assertEqual(
            pm.findall(') K'),
            self.expected_matches(((')', 1),('K', 6),)))
        self.assertEqual(
            pm.findall('} K'),
            self.expected_matches((('}', 3),('K', 6),)))
        self.assertEqual(
            pm.findall(' K('),
            self.expected_matches((('K', 6),('(', 0),))) # lex only here
        self.assertEqual(
            pm.findall(' K ('),
            self.expected_matches((('K', 6),('(', 0),))) # lex only here
        self.assertEqual(
            pm.findall(' K{'),
            self.expected_matches((('K', 6),('{', 2),)))
        self.assertEqual(
            pm.findall(' K {'),
            self.expected_matches((('K', 6),('{', 2),)))
        self.assertEqual(
            pm.findall(' input'),
            self.expected_matches((('input', 7),)))
        self.assertEqual(
            pm.findall(' shift'),
            self.expected_matches((('shift', 8),)))
        self.assertEqual(
            pm.findall(' wtm'),
            self.expected_matches((('wtm', 9),)))
        self.assertEqual(
            pm.findall(' lcamax'),
            self.expected_matches((('lcamax', 10),)))
        self.assertEqual(
            pm.findall(' from'),
            self.expected_matches((('from', 11),)))
        self.assertEqual(
            pm.findall(' 13'),
            self.expected_matches((('13', 12),)))
        self.assertEqual(
            pm.findall('"string"'),
            self.expected_matches((('"string"', 13),)))
        self.assertEqual(
            pm.findall(' or'),
            self.expected_matches((('or', 14),)))
        self.assertEqual(
            pm.findall(' not'),
            self.expected_matches((('not', 15),)))
        self.assertEqual(
            pm.findall(' attack'),
            self.expected_matches((('attack', 16),)))
        self.assertEqual(
            pm.findall(' between'),
            self.expected_matches((('between', 17),)))
        self.assertEqual(
            pm.findall(' power'),
            self.expected_matches((('power', 18),)))
        self.assertEqual(
            pm.findall(' elo'),
            self.expected_matches((('elo', 19),)))
        self.assertEqual(
            pm.findall(' comment'),
            self.expected_matches((('comment', 20),)))
        self.assertEqual(
            pm.findall(' hascomment'),
            self.expected_matches((('hascomment', 21),)))
        self.assertEqual(
            pm.findall(' mainline'),
            self.expected_matches((('mainline', 22),)))
        self.assertEqual(
            pm.findall(' move'),
            self.expected_matches((('move', 23),)))
        self.assertEqual(
            pm.findall(' movenumber'),
            self.expected_matches((('movenumber', 24),)))
        self.assertEqual(
            pm.findall(' next2'),
            self.expected_matches((('next2', 25),)))
        self.assertEqual(
            pm.findall(' on'),
            self.expected_matches((('on', 26),)))
        self.assertEqual(
            pm.findall(' player'),
            self.expected_matches((('player', 27),)))
        self.assertEqual(
            pm.findall(' powerdifference'),
            self.expected_matches((('powerdifference', 28),)))
        self.assertEqual(
            pm.findall(' ray'),
            self.expected_matches((('ray', 29),)))
        self.assertEqual(
            pm.findall(' relation'),
            self.expected_matches((('relation', 30),)))
        self.assertEqual(
            pm.findall(' result'),
            self.expected_matches((('result', 31),)))
        self.assertEqual(
            pm.findall(' sort'),
            self.expected_matches((('sort', 32),)))
        self.assertEqual(
            pm.findall(' year'),
            self.expected_matches((('year', 33),)))
        self.assertEqual(
            pm.findall(' silent'),
            self.expected_matches((('silent', 34),)))
        self.assertEqual(
            pm.findall(' event'),
            self.expected_matches((('event', 35),)))
        self.assertEqual(
            pm.findall(' piece'),
            self.expected_matches((('piece', 36),)))
        self.assertEqual(
            pm.findall(' in'),
            self.expected_matches((('in', 37),)))
        self.assertEqual(
            pm.findall(' origin'),
            self.expected_matches((('origin', 38),)))
        self.assertEqual(
            pm.findall(' $x '),
            self.expected_matches((('$x', 39),))) # Trailing [^a-zA-Z0-9] ends.
        self.assertEqual(
            pm.findall(' white'),
            self.expected_matches((('white', 40),)))
        self.assertEqual(
            pm.findall(' 1-0'),
            self.expected_matches((('1-0', 41),)))
        self.assertEqual(
            pm.findall('*'),
            self.expected_matches((('*', 42),)))
        self.assertEqual(
            pm.findall('qwerty'),
            self.expected_matches((('qwerty', 43),)))

        # Some which look like valid Chess Query Language
        self.assertEqual(
            pm.findall('\n'.join(('cql (',
                                  'input myfile.pgn',
                                  'output myfile_out.pgn',
                                  'gamenumber 100 25000',
                                  'year 1999 2013',
                                  'silent',
                                  'player "Jones"',
                                  'player white "Smith"',
                                  'player black "Wilson"',
                                  'elo 2200 2300',
                                  'elo white 1900 2000',
                                  'elo black 2050 2100',
                                  'site "London"',
                                  'event "4NCL"',
                                  'matchcount 1',
                                  'sort matchcount 1 3',
                                  ')'
                                  'Ra7',
                                  '{',
                                  '[Kk]a2-4',
                                  '}',
                                  'result 0-1',
                                  'result 1-0',
                                  'result 1/2-1/2',
                                  '; End of test',
                                  ))),
            self.expected_matches((('cql', 5),
                                   ('(', 0),
                                   ('input', 7),
                                   ('myfile.pgn', 41),
                                   ('output', 7),
                                   ('myfile_out.pgn', 41),
                                   ('gamenumber', 7),
                                   ('100', 12),
                                   ('25000', 12),
                                   ('year', 33),
                                   ('1999', 12),
                                   ('2013', 12),
                                   ('silent', 34),
                                   ('player', 27),
                                   ('"Jones"', 13),
                                   ('player', 27),
                                   ('white', 40),
                                   ('"Smith"', 13),
                                   ('player', 27),
                                   ('black', 40),
                                   ('"Wilson"', 13),
                                   ('elo', 19),
                                   ('2200', 12),
                                   ('2300', 12),
                                   ('elo', 19),
                                   ('white', 40),
                                   ('1900', 12),
                                   ('2000', 12),
                                   ('elo', 19),
                                   ('black', 40),
                                   ('2050', 12),
                                   ('2100', 12),
                                   ('site', 35),
                                   ('"London"', 13),
                                   ('event', 35),
                                   ('"4NCL"', 13),
                                   ('matchcount', 7),
                                   ('1', 12),
                                   ('sort', 32),
                                   ('matchcount', 7),
                                   ('1', 12),
                                   ('3', 12),
                                   (')', 1),
                                   ('Ra7', 6),
                                   ('{', 2),
                                   ('[Kk]a2-4', 6),
                                   ('}', 3),
                                   ('result', 31),
                                   ('0-1', 41),
                                   ('result', 31),
                                   ('1-0', 41),
                                   ('result', 31),
                                   ('1/2-1/2', 41),
                                   ('; End of test', 4),
                                   )))
        self.assertEqual(
            pm.findall('cql () wtm'),
            self.expected_matches((('cql', 5),
                                   ('(', 0),
                                   (')', 1),
                                   ('wtm', 9),
                                   )))
        self.assertEqual(
            pm.findall('cql () shift 1 3 k'),
            self.expected_matches((('cql', 5),
                                   ('(', 0),
                                   (')', 1),
                                   ('shift', 8),
                                   ('1', 12),
                                   ('3', 12),
                                   ('k', 6),
                                   )))
        self.assertEqual(
            pm.findall('cql() diagonal k or q'),
            self.expected_matches((('cql', 5),
                                   ('(', 0),
                                   (')', 1),
                                   ('diagonal', 8),
                                   ('k', 6),
                                   ('or', 14),
                                   ('q', 6),
                                   )))
        self.assertEqual(
            pm.findall('cql() not k'),
            self.expected_matches((('cql', 5),
                                   ('(', 0),
                                   (')', 1),
                                   ('not', 15),
                                   ('k', 6),
                                   )))
        self.assertEqual(
            pm.findall('cql(){}'),
            self.expected_matches((('cql', 5),
                                   ('(', 0),
                                   (')', 1),
                                   ('{', 2),
                                   ('}', 3),
                                   )))
        self.assertEqual(
            pm.findall('cql();comment'),
            self.expected_matches((('cql', 5),
                                   ('(', 0),
                                   (')', 1),
                                   (';comment', 4),
                                   )))
        self.assertEqual(
            pm.findall('cql()attack(Q r)'),
            self.expected_matches((('cql', 5),
                                   ('(', 0),
                                   (')', 1),
                                   ('attack', 16),
                                   ('(', 0),
                                   ('Q', 6),
                                   ('r', 6),
                                   (')', 1),
                                   )))
        self.assertEqual(
            pm.findall('cql() elo 2500'),
            self.expected_matches((('cql', 5),
                                   ('(', 0),
                                   (')', 1),
                                   ('elo', 19),
                                   ('2500', 12),
                                   )))
        self.assertEqual(
            pm.findall('cql() mainline'),
            self.expected_matches((('cql', 5),
                                   ('(', 0),
                                   (')', 1),
                                   ('mainline', 22),
                                   )))
        self.assertEqual(
            pm.findall('cql()mainline'),
            self.expected_matches((('cql', 5),
                                   ('(', 0),
                                   (')', 1),
                                   ('mainline', 22),
                                   )))
        self.assertEqual(
            pm.findall('cql() variation'),
            self.expected_matches((('cql', 5),
                                   ('(', 0),
                                   (')', 1),
                                   ('variation', 9)
                                   )))
        self.assertEqual(
            pm.findall('cql() move 20 35'),
            self.expected_matches((('cql', 5),
                                   ('(', 0),
                                   (')', 1),
                                   ('move', 23),
                                   ('20', 12),
                                   ('35', 12),
                                   )))
        

if __name__ == '__main__':
    runner = unittest.TextTestRunner
    loader = unittest.defaultTestLoader.loadTestsFromTestCase
    runner().run(loader(CQLConstants))




# piecedesignator.py
# Copyright 2017 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Chess Query Language (ChessQL) piece designator evaluator.

"""
import re
import collections

from .constants import (
    RANGE_SEPARATOR,
    COMPOUND_DESIGNATOR_START,
    COMPOUND_DESIGNATOR_END,
    ANY_WHITE_PIECE_NAME,
    ANY_BLACK_PIECE_NAME,
    PIECE_NAMES,
    FILE_RANGE,
    RANK_RANGE,
    FILE_DESIGNATOR,
    RANK_DESIGNATOR,
    SQUARE_DESIGNATOR_SEPARATOR,
    FILE_NAMES,
    RANK_NAMES,
    )


# Originally written to the cql1.0 definition.
# The composite piece types U, M, m, I, i, and ?, had disappeared by the cql5.0
# definition.
# The CQL syntax is very different but the piece designator syntax is the same,
# so some of this class needs to be re-written.
class PieceDesignator:
    """ChessQL piece designator base class.

    Impossible piece square combinations are accepted, pawns on ranks 1 and 8
    for example, but later processing may make assumptions about the need to
    look at the database for them.

    """

    # Assume string selected by cql.core.constants.PIECE_DESIGNATOR when using
    # these re elements, which have the same names as their counterparts.
    SIMPLE_SQUARE_DESIGNATOR = r''.join(
        (r'(', FILE_RANGE, RANK_RANGE, r')', r'|',
         r'(', FILE_RANGE, RANK_DESIGNATOR, r')', r'|',
         r'(', FILE_DESIGNATOR, RANK_RANGE, r')', r'|',
         r'(', FILE_DESIGNATOR, RANK_DESIGNATOR, r')'
         ))
    SQUARE_DESIGNATOR = r''.join(
        (r'(?:', COMPOUND_DESIGNATOR_START,
         r'([a-h][-a-h1-8]*[1-8](?:,[a-h][-a-h1-8]*[1-8])*)',
         COMPOUND_DESIGNATOR_END, r')', r'|',
         r'(?:', SIMPLE_SQUARE_DESIGNATOR, r')',
         ))
    PIECE_TYPE_DESIGNATOR = r''.join(
        (r'(?:', COMPOUND_DESIGNATOR_START,
         r'([^', COMPOUND_DESIGNATOR_END, r']+)',
         COMPOUND_DESIGNATOR_END, r')', r'|',
         r'([', PIECE_NAMES, r'])',
         ))
    PIECE_DESIGNATOR = r''.join(
        (r'(?:', SQUARE_DESIGNATOR, r')', r'|',
         r'(?:(?:(?:', PIECE_TYPE_DESIGNATOR,
         r')(?:', SQUARE_DESIGNATOR, r'))', r'|',
         r'(?:', PIECE_TYPE_DESIGNATOR, r')',
         r')',
         ))
    piece_designator = re.compile(PIECE_DESIGNATOR)

    PieceDesignator = collections.namedtuple(
        'PieceDesignator',
        ' '.join(('compoundsquare',
                  'filerankrange',
                  'filerange',
                  'rankrange',
                  'square',
                  'compoundpiece_s',
                  'piece_s',
                  'p_compoundsquare',
                  'p_filerankrange',
                  'p_filerange',
                  'p_rankrange',
                  'p_square',
                  'compoundpiece',
                  'piece',
                  )))

    # Default value.
    # When True the items in designator_set must not be in position.
    emptysquare = False

    def __init__(self, token):
        """"""
        self._token = token
        self._designator_set = None
        self._square_ranges_valid = None
        self._match = None
        self._groups = None

    def parse(self):
        """"""
        self._match = self.piece_designator.match(self._token)
        if self._match:
            self._groups = self.PieceDesignator(*self._match.groups(default=''))

    def __deepcopy__(self, memo):
        """Make a copy to which a transform can be applied.

        The transform is applied to the squares named in _groups and the piece
        designator text in _token.leaf is adjusted to fit.  Then _match and
        _groups are calculated using parse method.  These changes will be made
        to newcopy leaving self alone.

        """
        newcopy = empty_copy(self)
        newcopy._token = self._token
        newcopy._designator_set = None
        newcopy._square_ranges_valid = None
        newcopy._match = None
        newcopy._groups = self._groups
        return newcopy

    @property
    def designator_set(self):
        """"""
        return self._designator_set

    @staticmethod
    def _expand_composite_square(startfile, endfile, startrank, endrank):
        """"""
        files = FILE_NAMES[
            FILE_NAMES.index(startfile):FILE_NAMES.index(endfile)]
        files += endfile
        ranks = RANK_NAMES[
            RANK_NAMES.index(startrank):RANK_NAMES.index(endrank)]
        ranks += endrank
        return {f + r for f in files for r in ranks}

    def expand_piece_designator(self):
        """"""
        if self._groups is None:
            return
        squareset = set()
        self._square_ranges_valid = True
        for s in self.get_squares_list():
            if len(s) == 2:
                squareset.update(self._expand_composite_square(
                    s[0], s[0], s[1], s[1]))
            elif len(s) == 6:
                if s[0] > s[2] or s[3] > s[5]:
                    self._square_ranges_valid = False
                squareset.update(self._expand_composite_square(
                    s[0], s[2], s[3], s[5]))
            elif len(s) == 4:
                if s[1] == '-':
                    if s[0] > s[2]:
                        self._square_ranges_valid = False
                    squareset.update(self._expand_composite_square(
                        s[0], s[2], s[3], s[3]))
                elif s[2] == '-':
                    if s[1] > s[3]:
                        self._square_ranges_valid = False
                    squareset.update(self._expand_composite_square(
                        s[0], s[0], s[1], s[3]))
            else:
                squareset.add('') # Any square, s assumed to be ''.

        # Back to 'for p in self.get_pieces()' if 'all white pieces' and
        # 'all black pieces' indexes are created.  Similarely 'empty square',
        # which does not, and cannot, give correct answer yet.
        self._designator_set = {p + s
                                for p in self.get_pieces()
                                for s in squareset}

    @staticmethod
    def piece_square_to_index(designator_set, index_prefix):
        """Convert piece designator set values to index format: Qa4 to a4Q.

        Assumed that having all index values for a square adjacent is better
        than having index values for piece together, despite the need for
        conversion.

        """
        ds = set()
        for ps in designator_set:
            if len(ps) != 1:
                ds.add(index_prefix + ps)
            else:
                ds.update({index_prefix + ps + s
                           for s in PieceDesignator._expand_composite_square(
                               FILE_NAMES[0],
                               FILE_NAMES[-1],
                               RANK_NAMES[0],
                               RANK_NAMES[-1])})
        return ds

    # If 'square piece' is better order than 'piece square'
    @staticmethod
    def piece_square_to_index(designator_set, index_prefix):
        """Convert piece designator set values to index format: Qa4 to a4Q.

        Assumed that having all index values for a square adjacent is better
        than having index values for piece together, despite the need for
        conversion.

        """
        ds = set()
        for ps in designator_set:
            if len(ps) != 1:
                ds.add(index_prefix + ps[1:] + ps[0])
            else:
                ds.update({index_prefix + s + ps
                           for s in PieceDesignator._expand_composite_square(
                               FILE_NAMES[0],
                               FILE_NAMES[-1],
                               RANK_NAMES[0],
                               RANK_NAMES[-1])})
        return ds

    def get_shift_limits(self, ranklimits, filelimits):
        """Adjust ranges in ranklimits and filelimits to fit square designator.

        For first piece designator subject to one of the shift transforms the
        limits are ['8', '1'] and ['h', 'a'].

        The limits are not changed for piece designators which do not limit
        the squares ('q' for example).  Thus 'shift { Q q r }' would leave the
        limits as ['8', '1'] and ['h', 'a'].  The rank range '1-8' and the file
        range 'a-h' do not limit the squares in the relevant direction so these
        two ranges do not adjust the relevant limits.

        A piece designator with a single square sets the low and high limits to
        the square ( 'Kb2' produces ['2', '2'] and ['b', 'b'] for example).

        The four cases with rank and file ranges are:

        'Ra-h1-8' produces limits ['8', '1'] and ['h', 'a'] (not changed)
        'Ra-h2-8' produces limits ['2', '8'] and ['h', 'a']
        'Ra-d1-8' produces limits ['8', '1'] and ['a', 'd']
        'Ra-d2-8' produces limits ['2', '8'] and ['a', 'd']

        A piece designator with a compound square sets the limits to restrict
        shifting:

        'R[b2,g7}' produces ['2', '7'] and ['b', 'g']
        'R[a-h2,g7}' produces ['2', '7'] and ['g', 'g']
        'R[b1-8,g7}' produces ['7', '7'] and ['b', 'g']
        'R[a-h1-8,g7}' produces ['7', '7'] and ['g', 'g']

        Shifts on braced filters are similar:

        'shift { Rb2 kg7 }' produces ['2', '7'] and ['b', 'g']

        and so forth.

        shiftvertical and shifthorizontal generate the same limits but
        application to filters produces different results.

        """
        lowfile = FILE_NAMES[-1]
        highfile = FILE_NAMES[0]
        lowrank = RANK_NAMES[-1]
        highrank = RANK_NAMES[0]
        for s in self.get_squares_list():
            if len(s) == 2:
                lowfile = min(lowfile, s[0])
                highfile = max(highfile, s[0])
                lowrank = min(lowrank, s[1])
                highrank = max(highrank, s[1])
            elif len(s) == 6:
                if s[0] > s[2] or s[3] > s[5]:
                    self._square_ranges_valid = False
                if s[0] != FILE_NAMES[0] or s[2] != FILE_NAMES[-1]:
                    lowfile = min(lowfile, s[0])
                    highfile = max(highfile, s[2])
                if s[3] != RANK_NAMES[0] or s[5] != RANK_NAMES[-1]:
                    lowrank = min(lowrank, s[3])
                    highrank = max(highrank, s[5])
            elif len(s) == 4:
                if s[1] == RANGE_SEPARATOR:
                    if s[0] > s[2]:
                        self._square_ranges_valid = False
                    if (s[0] != FILE_NAMES[0] or
                        s[2] != FILE_NAMES[-1]):
                        lowfile = min(lowfile, s[0])
                        highfile = max(highfile, s[2])
                    lowrank = min(lowrank, s[3])
                    highrank = max(highrank, s[3])
                elif s[2] == RANGE_SEPARATOR:
                    if s[1] > s[3]:
                        self._square_ranges_valid = False
                    lowfile = min(lowfile, s[0])
                    highfile = max(highfile, s[0])
                    if (s[1] != RANK_NAMES[0] or
                        s[3] != RANK_NAMES[-1]):
                        lowrank = min(lowrank, s[1])
                        highrank = max(highrank, s[3])
        if ranklimits:
            ranklimits[0] = min(ranklimits[0], lowrank)
            ranklimits[1] = max(ranklimits[1], highrank)
        if filelimits:
            filelimits[0] = min(filelimits[0], lowfile)
            filelimits[1] = max(filelimits[1], highfile)

    def get_pieces(self):
        """"""
        g = self._groups
        pieces = ''.join((g.compoundpiece_s,
                          g.piece_s,
                          g.compoundpiece,
                          g.piece))
        if not pieces:
            pieces = ANY_WHITE_PIECE_NAME + ANY_BLACK_PIECE_NAME
        return pieces

    def get_squares(self):
        """"""
        g = self._groups
        return ''.join((g.compoundsquare,
                        g.filerankrange,
                        g.filerange,
                        g.rankrange,
                        g.square,
                        g.p_compoundsquare,
                        g.p_filerankrange,
                        g.p_filerange,
                        g.p_rankrange,
                        g.p_square))

    def get_squares_list(self):
        """"""
        return self.get_squares().split(SQUARE_DESIGNATOR_SEPARATOR)

    def is_compound_squares(self):
        """"""
        return self._groups.compoundsquare or self._groups.p_compoundsquare

    def is_compound_pieces(self):
        """"""
        return self._groups.compoundpiece_s or self._groups.compoundpiece


def empty_copy(obj):
    class Empty(obj.__class__):
        def __init__(self):
            pass
    newcopy = Empty()
    newcopy.__class__ = obj.__class__
    return newcopy



# __init__.py
# Copyright 2017 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Chess Query Language (CQL) parser.

The reference document used while writing this package is:

Chess Query Language documentation version 5.1 from http://www.gadycosteff.com

"""



# constants.py
# Copyright 2017 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Constants used when parsing Chess Query Language statements.
"""
import pgn_read.core.constants

RANGE_SEPARATOR = r'-'
COMPOUND_DESIGNATOR_START = r'\['
COMPOUND_DESIGNATOR_END = r'\]'
FILE_DESIGNATOR = r'[' + pgn_read.core.constants.FNR + r']'
RANK_DESIGNATOR = r'[' + pgn_read.core.constants.RNR + r']'
SQUARE_DESIGNATOR_SEPARATOR = r','
FILE_RANGE = FILE_DESIGNATOR + RANGE_SEPARATOR + FILE_DESIGNATOR
RANK_RANGE = RANK_DESIGNATOR + RANGE_SEPARATOR + RANK_DESIGNATOR
SIMPLE_SQUARE_DESIGNATOR = r''.join(
    (r'(?:', FILE_RANGE, RANK_RANGE, r')', r'|',
     r'(?:', FILE_RANGE, RANK_DESIGNATOR, r')', r'|',
     r'(?:', FILE_DESIGNATOR, RANK_RANGE, r')', r'|',
     r'(?:', FILE_DESIGNATOR, RANK_DESIGNATOR, r')'
     ))
SQUARE_DESIGNATOR = r''.join(
    (r'(?:', SIMPLE_SQUARE_DESIGNATOR, r')', r'|',
     r'(?:', COMPOUND_DESIGNATOR_START,
     r'(?:', SIMPLE_SQUARE_DESIGNATOR, r')',
     r'(?:', SQUARE_DESIGNATOR_SEPARATOR,
     r'(?:', SIMPLE_SQUARE_DESIGNATOR, r')', r')*',
     COMPOUND_DESIGNATOR_END, r')',
     ))

# Perhaps piece names should be imported from pgn_read package?
# See WKING so named to fit pgn_read.core.constants names.
WHITE_PIECE_NAMES = (
    pgn_read.core.constants.WKING +
    pgn_read.core.constants.WQUEEN +
    pgn_read.core.constants.WROOK +
    pgn_read.core.constants.WBISHOP +
    pgn_read.core.constants.WKNIGHT +
    pgn_read.core.constants.WPAWN)
BLACK_PIECE_NAMES = (
    pgn_read.core.constants.BKING +
    pgn_read.core.constants.BQUEEN +
    pgn_read.core.constants.BROOK +
    pgn_read.core.constants.BBISHOP +
    pgn_read.core.constants.BKNIGHT +
    pgn_read.core.constants.BPAWN)

ANY_WHITE_PIECE_NAME = r'A'
ANY_BLACK_PIECE_NAME = r'a'
EMPTY_SQUARE_NAME = r'.'
EMPTY_SQUARE_NAME_RE = r'\.'
ALL_GAMES_MATCH_PIECE_DESIGNATORS = (
    ANY_WHITE_PIECE_NAME +
    ANY_BLACK_PIECE_NAME +
    pgn_read.core.constants.WKING +
    pgn_read.core.constants.BKING +
    EMPTY_SQUARE_NAME)
ALL_PIECES = WHITE_PIECE_NAMES + BLACK_PIECE_NAMES + EMPTY_SQUARE_NAME

PIECE_NAMES = (WHITE_PIECE_NAMES +
               BLACK_PIECE_NAMES +
               ANY_WHITE_PIECE_NAME +
               ANY_BLACK_PIECE_NAME +
               EMPTY_SQUARE_NAME_RE)

PIECE_TYPE_DESIGNATOR = r''.join(
    (r'(?:[', PIECE_NAMES, r'])', r'|',
     r'(?:', COMPOUND_DESIGNATOR_START,
     r'[', PIECE_NAMES, r']+',
     COMPOUND_DESIGNATOR_END, r')',
     ))
PIECE_DESIGNATOR = r''.join(
    (r'(?:(?:(?:', PIECE_TYPE_DESIGNATOR,
     r')(?:', SQUARE_DESIGNATOR, r'))', r'|',
     r'(?:', SQUARE_DESIGNATOR, r')', r'|',
     r'(?:', PIECE_TYPE_DESIGNATOR, r')',
     r')',
     ))
PIECE_SQUARE_VARIABLE = r'\$[a-zA-Z0-9]+'

PIECE_DESIGNATOR_FILTER = 'piece_designator'

ANY = r'any'
ATTACK = r'attack'
BEGINVARIATION = r'beginvariation'
BETWEEN = r'between'
BTM = r'btm'
CHECK = r'check'
COMMENT = r'comment'
COUNTSQUARES = r'countsquares'
CQL = r'cql'
DARKSQUARES = r'darksquares'
# directions
UP = r'up'
DOWN = r'down'
LEFT = r'left' 
RIGHT = r'right'
NORTHEAST = r'northeast'
NORTHWEST = r'northwest'
SOUTHEAST = r'southeast'
SOUTHWEST = r'southwest'
VERTICAL = r'vertical'
HORIZONTAL = r'horizontal'
DIAGONAL = r'diagonal'
ORTHOGONAL = r'orthogonal'
ANYDIRECTION = r'anydirection'
# directions end
ELO = r'elo'
EVENT = r'event'
FLIP = r'flip'
FLIPDIHEDRAL = r'flipdihedral'
FLIPHORIZONTAL = r'fliphorizontal'
FLIPVERTICAL = r'flipvertical'
FLIPCOLOR = r'flipcolor'
GAMENUMBER = r'gamenumber'
HASCOMMENT = r'hascomment'
INITIAL = r'initial'
INPUT = r'input'
LIGHTSQUARES = r'lightsquares'
MAINLINE = r'mainline'
MATCHCOUNT = r'matchcount'
MATE = r'mate'
MOVE = r'move'
# move modifier
EMPTY = r'empty'
# move parameters
FROM = r'from'
TO = r'to'
PROMOTE = r'promote'
ENPASSANT = r'enpassant'
ENPASSANTSQUARE = r'enpassantsquare'
# move parameters end
MOVENUMBER = r'movenumber'
NEXT = r'next'
NEXT2 = r'next2'
NEXT_STAR = r'next\*'
NOT = r'not'
ON = r'on'
OR = r'or'
ORIGIN = r'origin'
OUTPUT = r'output'
PIECE = r'piece'
# piece parameters?
# piece parameters? end
PLAYER = r'player'
# player parameters
WHITE = r'white'
BLACK = r'black'
# player parameters end
POWER = r'power'
POWERDIFFERENCE = r'powerdifference'
PREVIOUS = r'previous'
PREVIOUS2 = r'previous2'
PREVIOUS_STAR = r'previous\*'
RAY = r'ray'
RELATION = r'relation'
# relation parameters
ANCESTOR = r'ancestor'
DESCENDANT = r'descendant'
ECHOFLIP = r'echoflip'
ECHOFLIPVERTICAL = r'echoflipvertical'
ECHOFLIPHORIZONTAL = r'echofliphorizontal'
ECHOSHIFT = r'echoshift'
ECHOSHIFTHORIZONTAL = r'echoshifthorizontal'
ECHOSHIFTVERTICAL = r'echoshiftvertical'
ECHOROTATE45 = r'echorotate45'
ECHOROTATE90 = r'echorotate90'
LCAMAX = r'lcamax'
LCASOURCE = r'lcasource'
LCASUBSTRING = r'lcasubstring'
LCASUM = r'lcasum'
LCATARGET = r'lcatarget'
MATCH = r'match'
MISMATCH = r'mismatch'
SOURCESQUARES = r'sourcesquares'
TARGETSQUARES = r'targetsquares'
TOMOVE = r'tomove'
# relation parameters end
RESULT = r'result'
ROTATE45 = r'rotate45'
ROTATE90 = r'rotate90'
SHIFT = r'shift'
SHIFTHORIZONTAL = r'shifthorizontal'
SHIFTVERTICAL = r'shiftvertical'
SILENT = r'silent'
SITE = r'site'
SORT = r'sort'
SQUARE = r'square'
# square parameters
ALL = r'all'
IN = r'in'
# square parameters end
STALEMATE = r'stalemate'
TERMINAL = r'terminal'
VARIATION = r'variation'
VARIATIONS = r'variations'
WTM = r'wtm'
YEAR = r'year'

# Mostly keywords are collated according to their use signature.
# Some are put together, separate from others with same signature, because
# they can appear in both the parameter and body parts of a statement.
# Those which appear just in the parameter part are put together despite their
# different use sigantures.
PIECE_SQUARE = r'|'.join(
    [f[-1] for f in sorted(((len(v), v) for v in {PIECE, SQUARE,}),
                           reverse=True)])
PIECE_SQUARE_KEYWORDS = r'|'.join(
    [f[-1] for f in sorted(((len(v), v) for v in {ALL, IN,}),
                           reverse=True)])
WHITE_BLACK_KEYWORDS = r'|'.join(
    [f[-1] for f in sorted(((len(v), v) for v in {WHITE, BLACK,}),
                           reverse=True)])
SITE_EVENT = r'|'.join(
    [f[-1] for f in sorted(((len(v), v) for v in {SITE, EVENT,}),
                           reverse=True)])
COUNTSQUARES_POWER = r'|'.join(
    [f[-1] for f in sorted(((len(v), v) for v in {COUNTSQUARES, POWER,}),
                           reverse=True)])
NEXT_PREVIOUS = r'|'.join(
    [f[-1] for f in sorted(((len(v), v) for v in {NEXT,
                                                  NEXT2,
                                                  PREVIOUS,
                                                  PREVIOUS2,}),
                           reverse=True)])
PLAIN = r'|'.join(
    [f[-1] for f in sorted(((len(v), v) for v in {ANY,
                                                  BEGINVARIATION,
                                                  BTM,
                                                  CHECK,
                                                  DARKSQUARES,
                                                  INITIAL,
                                                  LIGHTSQUARES,
                                                  MATE,
                                                  TERMINAL,
                                                  VARIATION,
                                                  STALEMATE,
                                                  WTM,}),
                           reverse=True)])
PLAIN_FILTER = 'plain'
RELATION_PARAMETER = r'|'.join(
    [f[-1] for f in sorted(((len(v), v) for v in {MATCH,
                                                  MISMATCH,
                                                  SOURCESQUARES,
                                                  TARGETSQUARES,
                                                  ANCESTOR,
                                                  DESCENDANT,
                                                  LCAMAX,
                                                  LCASOURCE,
                                                  LCASUBSTRING,
                                                  LCASUM,
                                                  LCATARGET,
                                                  ECHOFLIP,
                                                  ECHOFLIPHORIZONTAL,
                                                  ECHOFLIPVERTICAL,
                                                  ECHOROTATE45,
                                                  ECHOROTATE90,
                                                  ECHOSHIFT,
                                                  ECHOSHIFTHORIZONTAL,
                                                  ECHOSHIFTVERTICAL,
                                                  TOMOVE,}),
                           reverse=True)])
MOVE_PARAMETER = r'|'.join(
    [f[-1] for f in sorted(((len(v), v) for v in {EMPTY,
                                                  FROM,
                                                  TO,
                                                  PROMOTE,
                                                  ENPASSANT,
                                                  ENPASSANTSQUARE,}),
                           reverse=True)])
CQL_PARAMETER = r'|'.join(
    [f[-1] for f in sorted(((len(v), v) for v in {INPUT,
                                                  OUTPUT,
                                                  GAMENUMBER,
                                                  MATCHCOUNT,
                                                  VARIATIONS,}),
                           reverse=True)])
CYCLIC_FILTER = frozenset((UP,
                           DOWN,
                           LEFT,
                           RIGHT,
                           NORTHEAST,
                           NORTHWEST,
                           SOUTHEAST,
                           SOUTHWEST,
                           ))
DIRECTION_FILTER = frozenset(CYCLIC_FILTER.union((VERTICAL,
                                                  HORIZONTAL,
                                                  DIAGONAL,
                                                  ORTHOGONAL,
                                                  ANYDIRECTION,
                                                  )))
DIHEDRAL_FILTER = frozenset((FLIP,
                             FLIPDIHEDRAL,
                             FLIPHORIZONTAL,
                             FLIPVERTICAL,
                             ROTATE90,
                             ))
SHIFT_FILTER = frozenset((SHIFT,
                          SHIFTHORIZONTAL,
                          SHIFTVERTICAL,
                          ))
TRANSFORM = r'|'.join(
    [f[-1] for f in sorted(((len(v), v)
                            for v in DIRECTION_FILTER.union(
                                DIHEDRAL_FILTER.union(
                                    SHIFT_FILTER.union(
                                        {FLIPCOLOR,
                                         NEXT_STAR,
                                         PREVIOUS_STAR,
                                         ROTATE45,
                                         })))),
                           reverse=True)])

# Used as the type of a node in a statement node tree: not to be confused with
# LEFT_PARENTHESIS and LEFT_BRACE defined for use in regular expressions.
LEFT_PARENTHESIS_FILTER = '('
LEFT_BRACE_FILTER = '{'
RELATION_PARAMETER_NAME = 'relation_parameter'
RIGHT_PARENTHESIS_TYPE = ')'
PGN_FILE_VALUE = 'pgn'

# Any filter not in this frozenset is definitely not a set filter.
# Left parenthesis and left brace are included because they are used as the
# node type in the node tree for a statement, and their status depends on the
# enclosed filters.
SET_FILTER = frozenset(
    DIRECTION_FILTER.union(
        DIHEDRAL_FILTER.union(
            SHIFT_FILTER.union({ATTACK,
                                MOVE,
                                DARKSQUARES,
                                LIGHTSQUARES,
                                PIECE,
                                SQUARE,
                                NOT,
                                ON,
                                OR,
                                BETWEEN,
                                PIECE_DESIGNATOR_FILTER,
                                LEFT_BRACE_FILTER,
                                LEFT_PARENTHESIS_FILTER,
                                }))))
                                              

LEFT_PARENTHESIS = r'\('
RIGHT_PARENTHESIS = r'\)'
LEFT_BRACE = r'\{'
RIGHT_BRACE = r'\}'
NUMBER = r'\d+'
DOUBLE_QUOTED_STRING = r'"[^\\"]*(?:\\.[^\\"]*)*"'
COMMENT_STRING = r';[^\n]*'

# PGN allows '*' as a result but CQL does not.
# CQL states filenames must end in the extension '.pgn', and cql5.1 rejects
# commands starting like:
# 'cql ( input my games.pgn )' but accepts:
# 'cql ( input mygames.pgn )'.  Spaces are not allowed in filenames.
PGN_FILE = r'\.pgn'
ALLOWED_STRINGS = r''.join((r'(?:',
                            pgn_read.core.constants.WHITE_WIN,
                            r')|(?:',
                            pgn_read.core.constants.BLACK_WIN,
                            r')|(?:',
                            pgn_read.core.constants.DRAW,
                            r')|(?:[^\s]+',
                            PGN_FILE,
                            r')',
                            ))

# These are used in some CQL examples but documentation is not clear on rules.
# Looks similar to re module usage.
# '*' also appears as the trailing character of 'next*' and 'previous*'.
ZERO_OR_ONE = r'\?'
ZERO_OR_MORE = r'\*'
ONE_OR_MORE = r'\+'

# REPEAT_OPERATORS must appear in CQL_PATTERN after COMMENT_STRING, NEXT_STAR,
# PREVIOUS_STAR, DOUBLE_QUOTED_STRING, and ALLOWED_STRINGS, to avoid treating
# a '*' embedded in something else incorrectly.
# Best to include PIECE_DESIGNATOR in this list although there is no conflict
# of interest at present.
REPEAT_OPERATORS = r'|'.join((ZERO_OR_ONE, ZERO_OR_MORE, ONE_OR_MORE))

BAD_TOKEN = r'\S+'

EDGING = r'|'.join((r'\s',
                    LEFT_PARENTHESIS,
                    RIGHT_PARENTHESIS,
                    LEFT_BRACE,
                    RIGHT_BRACE))
REPEAT_EDGING = r'|'.join((r'\s',
                           LEFT_PARENTHESIS,
                           RIGHT_PARENTHESIS,
                           LEFT_BRACE,
                           RIGHT_BRACE,
                           REPEAT_OPERATORS,
                           ))
LEADING = EDGING.join((r'(?<=', r')'))
TRAILING = EDGING.join((r'(?=', r'|\Z)'))
#REPEAT_LEADING = REPEAT_EDGING.join((r'(?<=', r')'))
REPEAT_TRAILING = REPEAT_EDGING.join((r'(?=', r'|\Z)'))

CQL_PATTERN = r'|'.join((LEFT_PARENTHESIS.join((r'(', r')')),
                         RIGHT_PARENTHESIS.join((r'(', r')')),
                         LEFT_BRACE.join((r'(', r')')),
                         RIGHT_BRACE.join((r'(', r')')),
                         COMMENT_STRING.join((r'(', r')')),
                         CQL.join((r'(', r')')).join(
                             (r'(?:\A\s*)',
                              LEFT_PARENTHESIS.join((r'(?=\s|', r')')))),
                         PIECE_DESIGNATOR.join(
                             (r'(', r')')).join((LEADING, REPEAT_TRAILING)),
                         CQL_PARAMETER.join(
                             (r'(', r')')).join((LEADING, TRAILING)),
                         TRANSFORM.join(
                             (r'(', r')')).join((LEADING, TRAILING)),
                         PLAIN.join(
                             (r'(', r')')).join((LEADING, REPEAT_TRAILING)),
                         RELATION_PARAMETER.join(
                             (r'(', r')')).join((LEADING, TRAILING)),
                         MOVE_PARAMETER.join(
                             (r'(', r')')).join((LEADING, TRAILING)),
                         NUMBER.join(
                             (r'(', r')')).join((LEADING, REPEAT_TRAILING)),
                         DOUBLE_QUOTED_STRING.join((r'(', r')')),
                         OR.join((r'(', r')')).join((LEADING, TRAILING)),
                         NOT.join((r'(', r')')).join((LEADING, TRAILING)),
                         ATTACK.join((r'(', r')')).join(
                             (LEADING, TRAILING)), # CQL body or ray keyword.
                         BETWEEN.join((r'(', r')')).join((LEADING, TRAILING)),
                         COUNTSQUARES_POWER.join(
                             (r'(', r')')).join((LEADING, TRAILING)),
                         ELO.join((r'(', r')')).join(
                             (LEADING, TRAILING)), # CQL parameter or CQL body.
                         COMMENT.join((r'(', r')')).join((LEADING, TRAILING)),
                         HASCOMMENT.join(
                             (r'(', r')')).join((LEADING, TRAILING)),
                         MAINLINE.join((r'(', r')')).join(
                             (LEADING, TRAILING)), # CQL body or move keyword.
                         MOVE.join((r'(', r')')).join((LEADING, TRAILING)),
                         MOVENUMBER.join(
                             (r'(', r')')).join((LEADING, TRAILING)),
                         NEXT_PREVIOUS.join((r'(', r')')).join(
                             (LEADING, TRAILING)), # CQL body or move keyword.
                         ON.join((r'(', r')')).join((LEADING, TRAILING)),
                         PLAYER.join((r'(', r')')).join(
                             (LEADING, TRAILING)), # CQL parameter or CQL body.
                         POWERDIFFERENCE.join(
                             (r'(', r')')).join((LEADING, TRAILING)),
                         RAY.join((r'(', r')')).join((LEADING, TRAILING)),
                         RELATION.join((r'(', r')')).join((LEADING, TRAILING)),
                         RESULT.join((r'(', r')')).join(
                             (LEADING, TRAILING)), # CQL parameter or CQL body.
                         SORT.join((r'(', r')')).join(
                             (LEADING, TRAILING)), # CQL parameter or CQL body.
                         YEAR.join((r'(', r')')).join(
                             (LEADING, TRAILING)), # CQL parameter or CQL body.
                         SILENT.join((r'(', r')')).join(
                             (LEADING, TRAILING)), # CQL parameter or CQL body.
                         SITE_EVENT.join((r'(', r')')).join(
                             (LEADING, TRAILING)), # CQL parameter or CQL body.
                         PIECE_SQUARE.join(
                             (r'(', r')')).join((LEADING, TRAILING)),
                         PIECE_SQUARE_KEYWORDS.join((r'(', r')')).join(
                             (LEADING, TRAILING)), # piece or square keywords.
                         ORIGIN.join((r'(', r')')).join(
                             (LEADING, TRAILING)), # piece or square keywords.
                         PIECE_SQUARE_VARIABLE.join((r'(', r')')).join(
                             (LEADING, REPEAT_TRAILING)),
                         WHITE_BLACK_KEYWORDS.join((r'(', r')')).join(
                             (LEADING, TRAILING)), # player or elo keywords.
                         ALLOWED_STRINGS.join(
                             (r'(', r')')), # result, or *.pgn filename.
                         REPEAT_OPERATORS.join((r'(', r')')),
                         BAD_TOKEN.join((r'(', r')')),
                         ))

# The Python names of group items found by the regular expression pattern in
# CQL_PATTERN: the lower case version of the attribute name except where not
# possible due to Python syntax.
TOKEN_NAMES = ' '.join(('LEFT_PARENTHESIS'.lower(),
                        'RIGHT_PARENTHESIS'.lower(),
                        'LEFT_BRACE'.lower(),
                        'RIGHT_BRACE'.lower(),
                        'COMMENT_STRING'.lower(),
                        'CQL'.lower(),
                        'PIECE_DESIGNATOR'.lower(),
                        'CQL_PARAMETER'.lower(),
                        'TRANSFORM'.lower(),
                        'PLAIN'.lower(),
                        'RELATION_PARAMETER'.lower(),
                        'MOVE_PARAMETER'.lower(),
                        'NUMBER'.lower(),
                        'DOUBLE_QUOTED_STRING'.lower(),
                        'or_', #'OR'.lower(),
                        'not_', #'NOT'.lower(),
                        'ATTACK'.lower(),
                        'BETWEEN'.lower(),
                        'COUNTSQUARES_POWER'.lower(),
                        'ELO'.lower(),
                        'COMMENT'.lower(),
                        'HASCOMMENT'.lower(),
                        'MAINLINE'.lower(),
                        'MOVE'.lower(),
                        'MOVENUMBER'.lower(),
                        'NEXT_PREVIOUS'.lower(),
                        'ON'.lower(),
                        'PLAYER'.lower(),
                        'POWERDIFFERENCE'.lower(),
                        'RAY'.lower(),
                        'RELATION'.lower(),
                        'RESULT'.lower(),
                        'SORT'.lower(),
                        'YEAR'.lower(),
                        'SILENT'.lower(),
                        'SITE_EVENT'.lower(),
                        'PIECE_SQUARE'.lower(),
                        'PIECE_SQUARE_KEYWORDS'.lower(),
                        'ORIGIN'.lower(),
                        'PIECE_SQUARE_VARIABLE'.lower(),
                        'WHITE_BLACK_KEYWORDS'.lower(),
                        'ALLOWED_STRINGS'.lower(),
                        'REPEAT_OPERATORS'.lower(),
                        'BAD_TOKEN'.lower(),
                        ))
(RIGHT_PARENTHESIS_INDEX,
 LEFT_BRACE_INDEX,
 RIGHT_BRACE_INDEX,
 PIECE_DESIGNATOR_INDEX,
 PLAIN_INDEX,
 OR_INDEX,
 NOT_INDEX,
 ON_INDEX,
 ) = [
    e for e, t in enumerate(TOKEN_NAMES.split())
    if t in {'right_parenthesis',
             'left_brace',
             'right_brace',
             'piece_designator',
             'plain',
             'or_',
             'not_',
             'on',
             }]

# CQL_PATTERN and TOKEN_NAMES do not use the concept a 'range' is a sequence of
# 1 or 2 'numbers'.  The presence or absence of a 'range' can determine whether
# a filter is a set filter (some filters are not set filters anyway).  This
# matters to 'transform_filter's at least, and the statement module uses RANGE
# to identify 'number's used as ranges.
RANGE = 'range'

# The CQL statement may be preceded by a statement name.
NAME_DELIMITER = '\n'

# File and rank names for piece designators.
FILE_NAMES = ''.join(tuple(sorted(pgn_read.core.constants.MAPFILE)))
RANK_NAMES = ''.join(tuple(sorted(pgn_read.core.constants.MAPRANK)))

# Derive all lists of squares in rays from pgn_read.core.constants.GAPS, where
# GAPS excludes the squares at each end of the gap but these are included in a
# ray when distinct.  Each gap is represented by a bitarray of square numbers,
# but ray processing needs an ordered list of square identities.
mfosn = {v:k
         for k, v
         in pgn_read.core.constants.MAP_PGN_SQUARE_NAME_TO_FEN_ORDER.items()}
RAYS = {}
for ef, f in enumerate(pgn_read.core.constants.GAPS):
    for et, t in enumerate(f):
        if t == pgn_read.core.constants.ALL_SQUARES or t == 0:
            continue
        fr = RAYS.setdefault(mfosn[ef], {})
        ls = [mfosn[ef]]
        fr[mfosn[et]] = ls
        step = 1 if et > ef else -1
        for s in range(ef + step, et, step):
            if pgn_read.core.constants.SQUARE_BITS[s] & t:
                ls.append(mfosn[s])
        ls.append(mfosn[et])

        # Generate rays of length 2 away from edge of board.
        if len(ls) == 3:
            RAYS[ls[0]][ls[1]] = ls[:2]

# There are no rays of length 3 going right from g4 and so forth, but the rays
# going left from h4 and so forth do exist.
supp = []
for fk in RAYS:
    for tv in RAYS[fk].values():
        if len(tv) == 2:
            if tv[0] not in RAYS[tv[1]]:
                supp.append(tv)
for t, f in supp:
    RAYS[f][t] = [f, t]
del ef, f, et, t, fr, mfosn, ls, s, step, supp

# The diagonals of length 2 do not appear as part of a ray of length 3.
RAYS['a2']['b1'] = ['a2', 'b1']
RAYS['a7']['b8'] = ['a7', 'b8']
RAYS['b1']['a2'] = ['b1', 'a2']
RAYS['b8']['a7'] = ['b8', 'a7']
RAYS['h2']['g1'] = ['h2', 'g1']
RAYS['h7']['g8'] = ['h7', 'g8']
RAYS['g1']['h2'] = ['g1', 'h2']
RAYS['g8']['h7'] = ['g8', 'h7']
